from asyncio import Task, create_task, sleep
from collections import defaultdict
from enum import IntEnum
from functools import cached_property
from itertools import filterfalse
from logging import DEBUG, getLogger
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import a_sync
import brownie
from a_sync import igather
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import ZERO_ADDRESS
from brownie.convert.datatypes import EthAddress
from brownie.exceptions import ContractNotFound, EventLookupError
from brownie.network.event import _EventItem
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20, WeiBalance, _EventsLoader, _Loader
from y.constants import CHAINID, CONNECTED_TO_MAINNET
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import (
    Address,
    AddressOrContract,
    AnyAddressType,
    Block,
    Pool,
    UsdPrice,
    UsdValue,
)
from y.exceptions import (
    ContractNotVerified,
    MessedUpBrownieContract,
    PriceError,
    UnsupportedNetwork,
    yPriceMagicError,
    call_reverted,
)
from y.interfaces.curve.CurveRegistry import CURVE_REGISTRY_ABI
from y.networks import Network
from y.utils import a_sync_ttl_cache
from y.utils.events import ProcessedEvents
from y.utils.multicall import multicall_same_func_same_contract_different_inputs
from y.utils.raw_calls import raw_call

T = TypeVar("T")

logger = getLogger(__name__)
startup_logger = logger.getChild("startup")
_startup_logger_debug = startup_logger.debug

# curve registry documentation https://curve.readthedocs.io/registry-address-provider.html
ADDRESS_PROVIDER = "0x0000000022D53366457F9d5E68Ec105046FC4383"
"Curve's address provider contract on all chains."

DED_POOLS = {
    Network.Mainnet: {
        "0x8301AE4fc9c624d1D396cbDAa1ed877821D7C511": "0xEd4064f376cB8d68F770FB1Ff088a3d0F3FF5c4d",
    },
}.get(CHAINID, {})
"The on chain registry no longer returns the lp token address for these dead pools, so we need to provide it manually."


class Ids(IntEnum):
    Main_Registry = 0
    PoolInfo_Getters = 1
    Exchanges = 2
    Metapool_Factory = 3
    Fee_Distributor = 4
    CryptoSwap_Registry = 5
    CryptoPool_Factory = 6
    # On Mainnet, id 7 is listed as "metafactory".
    # On Polygon, id 7 is listed as "cryptopool factory".
    # On other chains, "cryptopool factory" is id 6.
    # On Polygon, id 6 is "crypto factory".
    # I've only seen this on Mainnet and Polygon so far, for now will treat `7` == `6`.
    Cryptopool_Factory = 7
    crvUSD_Plain_Pools_deprecated_1 = 8
    crvUSD_Plain_Pools_deprecated_2 = 9
    crvUSD_Plain_Pools = 10
    Curve_Tricrypto_Factory = 11
    CurveStableswapFactoryNG = 12


_METAPOOL_FACTORY_IDS = (
    Ids.Metapool_Factory,
    Ids.crvUSD_Plain_Pools,
    Ids.Curve_Tricrypto_Factory,
    Ids.CurveStableswapFactoryNG,
)

_LT = TypeVar("_LT", bound=_Loader)


class _CurveEventsLoader(_EventsLoader):
    _events: "CurveEvents"


class CurveEvents(ProcessedEvents[_EventItem]):
    __slots__ = ("_base",)

    def __init__(self, base: _LT):
        super().__init__(addresses=base.address)
        self._base = base


class AddressProviderEvents(CurveEvents):
    @property
    def provider(self) -> "AddressProvider":
        return self._base

    def _process_event(self, event) -> None:
        if event.name == "NewAddressIdentifier" and event["addr"] != ZERO_ADDRESS:
            self.provider.identifiers[Ids(event["id"])].append(event["addr"])
        elif event.name == "AddressModified" and event["new_address"] != ZERO_ADDRESS:
            self.provider.identifiers[Ids(event["id"])].append(event["new_address"])
        _startup_logger_debug("%s loaded event %s at block %s", self, event, event.block_number)
        return event


class RegistryEvents(CurveEvents):
    __slots__ = ("_tasks",)

    def __init__(self, base: _LT):
        super().__init__(base)
        self._tasks: List["Task[EthAddress]"] = []

    @property
    def registry(self) -> "Registry":
        return self._base

    def _process_event(self, event: _EventItem) -> None:
        if event.name == "PoolAdded":
            # TODO async this
            try:
                pool = event["pool"]
            except EventLookupError:
                pool = event["newPool"]
            self._tasks.append(
                create_task(
                    coro=self._add_pool(pool),
                    name=f"Registry._add_pool for pool {pool}",
                )
            )
            curve.registries[event.address].add(pool)
        elif event.name == "PoolRemoved":
            curve.registries[event.address].discard(event["pool"])
        _startup_logger_debug("%s loaded event %s at block %s", self, event, event.block_number)
        return event

    async def _add_pool(self, pool: Address) -> EthAddress:
        if pool in DED_POOLS:
            # The on chain registry no longer returns the lp token address for these dead pools, so we need to provide it manually.
            lp_token = DED_POOLS[pool]
        else:
            lp_token = await self.registry.contract.get_lp_token.coroutine(pool)
        curve.token_to_pool[lp_token] = pool

    async def _set_lock(self, block: int) -> None:
        await igather(self._tasks)
        self._tasks.clear()
        self._lock.set(block)


class AddressProvider(_CurveEventsLoader):
    __slots__ = (
        "identifiers",
        "_events",
    )

    def __init__(self, address: Address, *, asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self.identifiers = defaultdict(list)
        self._events = AddressProviderEvents(self)

    async def get_registry(self) -> EthAddress:
        contract = await Contract.coroutine(self.address)
        return await contract.get_registry

    async def _load_factories(self) -> None:
        # factory events are quite useless, so we use a different method
        if debug_logs := startup_logger.isEnabledFor(DEBUG):
            _startup_logger_log_debug("loading pools from metapool factories")
        # TODO: remove this once curve adds to address provider
        if CONNECTED_TO_MAINNET:
            self.identifiers[Ids.CurveStableswapFactoryNG] = [
                "0x6A8cbed756804B16E05E741eDaBd5cB544AE21bf"
            ]

        if metapool_factories := [
            Factory(factory, asynchronous=self.asynchronous)
            for factories in map(self.identifiers.get, _METAPOOL_FACTORY_IDS)
            for factory in factories or ()
        ]:
            async for factory, pool_list in a_sync.map(Factory.read_pools, metapool_factories):
                for pool in pool_list:
                    # for metpool factories pool is the same as lp token
                    curve.token_to_pool[pool] = pool
                    curve.factories[factory].add(pool)

        # if there are factories that haven't yet been added to the on-chain address provider,
        # please refer to commit 3f70c4246615017d87602e03272b3ed18d594d3c to see how to add them manually
        if (
            identifiers := self.identifiers[Ids.CryptoPool_Factory]
            + self.identifiers[Ids.Cryptopool_Factory]
        ):
            if debug_logs:
                _startup_logger_log_debug("loading pools from cryptopool factories")
            await a_sync.map(Factory, identifiers, asynchronous=self.asynchronous)

        if not curve._done.is_set():
            startup_logger.info(
                "loaded %s pools from %s registries and %s factories",
                len(curve.token_to_pool),
                len(curve.registries),
                len(curve.factories),
            )
            curve._done.set()


class Registry(_CurveEventsLoader):
    __slots__ = "_events"

    def __init__(self, address: Address, curve: "CurveRegistry", *, asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self._events = RegistryEvents(self)


class Factory(_Loader):
    async def get_pool(self, i: int) -> EthAddress:
        contract = await Contract.coroutine(self.address)
        return await contract.pool_list.coroutine(i)

    async def pool_count(self, block: Optional[int] = None) -> int:
        contract = await Contract.coroutine(self.address)
        return await contract.pool_count.coroutine(block_identifier=block)

    async def read_pools(self) -> List[EthAddress]:
        try:
            # lets load the contract async and then we can use the sync property more conveniently
            await Contract.coroutine(self.address)
        except ContractNotVerified:
            if CHAINID == Network.xDai:
                Contract.from_abi("Vyper_contract", self.address, CURVE_REGISTRY_ABI)
            else:
                # This happens sometimes, not sure why as the contract is verified.
                brownie.Contract.from_explorer(self.address)
        pool_count = await self.pool_count()
        return await a_sync.map(self.get_pool, range(pool_count)).values(pop=True)

    async def _load(self) -> None:
        pool_list = await self.read_pools(sync=False)
        debug_logs = startup_logger.isEnabledFor(DEBUG)
        await igather(
            self.__load_pool(pool, debug_logs)
            for pool in pool_list
            if pool not in curve.factories[self.address]
        )
        self._loaded.set()
        if debug_logs:
            _startup_logger_log_debug("loaded %s pools for %s", len(pool_list), self)

    async def __load_pool(self, pool: Address, debug_logs: bool) -> None:
        factory = await Contract.coroutine(self.address)
        # for curve v5 pools, pool and lp token are separate
        if hasattr(factory, "get_token"):
            lp_token = await factory.get_token.coroutine(pool)
        elif hasattr(factory, "get_lp_token"):
            lp_token = await factory.get_lp_token.coroutine(pool)
        else:
            raise NotImplementedError(
                f"New factory {factory.address} is not yet supported. Please notify a ypricemagic maintainer."
            )
        curve.token_to_pool[lp_token] = pool
        curve.factories[factory.address].add(pool)
        if debug_logs:
            _startup_logger_log_debug("loaded %s pool %s", self, pool)


class CurvePool(ERC20):
    """
    Represents a Curve pool.

    This class provides methods to interact with Curve pools, including fetching
    pool information, calculating token exchanges, and retrieving liquidity data.

    Note:
        This class inherits from :class:`~y.classes.common.ERC20` for convenience,
        but a Curve pool is not always an ERC20 token. This inheritance is used
        to leverage existing functionality for token interactions.

    Methods:
        - :meth:`coins`: Fetches the coins in the pool.
        - :meth:`get_dy`: Calculates the amount of output tokens for a given input amount, simulating a token exchange.
        - :meth:`get_balances`: Retrieves the balances of tokens in the pool.
        - :meth:`get_tvl`: Retrieves the total value locked in the pool.

    Examples:
        >>> pool = CurvePool("0x1234567890abcdef1234567890abcdef12345678")
        >>> await pool.coins
        [<ERC20 TKN1 '0x...'>, <ERC20 TKN2 '0x...'>]
    """

    @a_sync.aka.cached_property
    async def factory(self) -> Contract:
        return await curve.get_factory(self, sync=False)

    __factory__: HiddenMethodDescriptor[Self, Contract]

    @a_sync.aka.cached_property
    async def coins(self) -> List[ERC20]:
        """
        Get coins of pool.

        Returns:
            A list of :class:`~y.classes.common.ERC20` tokens representing the coins in the pool.

        Examples:
            >>> pool = CurvePool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.coins
            [<ERC20 TKN1 '0x...'>, <ERC20 TKN2 '0x...'>]
        """
        factory = await self.__factory__
        if factory:
            coins = await factory.get_coins.coroutine(self.address)
        else:
            registry = await curve.__registry__
            coins = await registry.get_coins.coroutine(self.address)

        # pool not in registry
        if set(coins) == {ZERO_ADDRESS}:
            coins = await multicall_same_func_same_contract_different_inputs(
                self.address,
                "coins(uint256)(address)",
                inputs=list(range(8)),
                return_None_on_failure=True,
                sync=False,
            )

        return [
            ERC20(coin, asynchronous=self.asynchronous)
            for coin in coins
            if coin not in {None, ZERO_ADDRESS}
        ]

    __coins__: HiddenMethodDescriptor[Self, List[ERC20]]

    @a_sync.a_sync(ram_cache_maxsize=256)
    async def get_coin_index(self, coin: AnyAddressType) -> int:
        return [i for i, _coin in enumerate(await self.__coins__) if _coin == coin][0]

    @a_sync.aka.cached_property
    async def num_coins(self) -> int:
        return len(await self.__coins__)

    __num_coins__: HiddenMethodDescriptor[Self, int]

    async def get_dy(
        self,
        coin_ix_in: int,
        coin_ix_out: int,
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[WeiBalance]:
        tokens = await self.__coins__
        token_in: ERC20 = tokens[coin_ix_in]
        token_out: ERC20 = tokens[coin_ix_out]
        amount_in = await token_in.__scale__
        contract = await Contract.coroutine(self.address)
        try:
            amount_out = await contract.get_dy.coroutine(
                coin_ix_in, coin_ix_out, amount_in, block_identifier=block
            )
            return WeiBalance(
                amount_out,
                token_out,
                block=block,
                ignore_pools=(*ignore_pools, self),
                skip_cache=skip_cache,
            )
        except Exception as e:
            if call_reverted(e):
                return None
            raise

    @a_sync.aka.cached_property
    async def coins_decimals(self) -> List[int]:
        factory = await self.__factory__
        source = factory or await curve.registry
        coins_decimals = await source.get_decimals.coroutine(self.address)

        # pool not in registry
        if not any(coins_decimals):
            coins_decimals = await a_sync.map(ERC20.decimals, await self.__coins__).values(pop=True)

        return [dec for dec in coins_decimals if dec != 0]

    __coins_decimals__: HiddenMethodDescriptor[Self, List[int]]

    @a_sync.aka.cached_property
    async def get_underlying_coins(self) -> List[ERC20]:
        factory = await self.__factory__
        if factory:
            # new factory reverts for non-meta pools
            if not hasattr(factory, "is_meta") or factory.is_meta(self.address):
                coins = await factory.get_underlying_coins.coroutine(self.address)
            else:
                coins = await factory.get_coins.coroutine(self.address)
        else:
            registry = await curve.registry
            coins = await registry.get_underlying_coins.coroutine(self.address)

        # pool not in registry, not checking for underlying_coins here
        if set(coins) == {ZERO_ADDRESS}:
            return await self.__coins__

        return [
            ERC20(coin, asynchronous=self.asynchronous) for coin in coins if coin != ZERO_ADDRESS
        ]

    __get_underlying_coins__: HiddenMethodDescriptor[Self, List[ERC20]]

    @a_sync.a_sync(ram_cache_maxsize=1000)
    async def get_balances(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> List[WeiBalance]:
        """
        Get {token: balance} of liquidity in the pool.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: If True, skip using the cache while fetching balance data.

        Returns:
            A list of :class:`~y.classes.common.WeiBalance` objects representing the balances of tokens in the pool.

        Examples:
            >>> pool = CurvePool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.get_balances()
            [<WeiBalance token=<ERC20 TKN1 '0x...'> balance=1000000000000000000 block=None>, ...]
        """
        coins = await self.__coins__
        try:
            factory = await self.__factory__
            source = factory or await curve.__registry__
            balances = await source.get_balances.coroutine(self.address, block_identifier=block)
        except (ContractLogicError, ValueError):
            # ContractLogicError in web3>=6.0, ValueError in <6.0
            # fallback for historical queries where registry was not yet deployed
            balances = await a_sync.map(self._get_balance, range(len(coins)), block=block).values(
                pop=True
            )

        if not any(balances):
            raise ValueError(f"could not fetch balances {self.__str__()} at {block}")

        return [
            WeiBalance(balance, coin, block, skip_cache=skip_cache)
            for coin, balance in zip(coins, balances)
            if coin != ZERO_ADDRESS
        ]

    async def _get_balance(self, i: int, block: Optional[Block] = None) -> Optional[int]:
        try:
            contract = await Contract.coroutine(self.address)
        except ContractNotVerified:
            # TODO: figure out if we need to build this, usually they get verified quickly
            return None
        try:
            return await contract.balances.coroutine(i, block_identifier=block)
        except ContractLogicError as e:
            # happens on web3py>=6.0
            if str(e) == "execution reverted":
                return None
            raise
        except ValueError as e:
            # happens on web3py<6.0
            if str(e) == "No data was returned - the call likely reverted":
                return None
            raise

    async def get_tvl(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Optional[UsdValue]:
        """
        Get total value in Curve pool.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: If True, skip using the cache while fetching TVL data.

        Returns:
            The total value locked in the pool as a :class:`~y.datatypes.UsdValue`, or None if the TVL cannot be determined.

        Examples:
            >>> pool = CurvePool("0x1234567890abcdef1234567890abcdef12345678")
            >>> await pool.get_tvl()
            UsdValue(1234567.89)
        """
        try:
            price = await WeiBalance.value_usd.sum(
                self.get_balances(block=block, skip_cache=skip_cache, sync=False),
                sync=False,
            )
            return UsdValue(price)
        except ValueError as e:
            if str(e).startswith("could not fetch balances "):
                logger.debug("could not fetch balances for %s", self)
                return None
            raise
        except yPriceMagicError as e:
            if not isinstance(e.exception, PriceError):
                raise
            logger.debug("%s in %s", repr(e), self)
            return None

    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60 * 60)
    async def check_liquidity(self, token: Address, block: Block) -> int:
        if block < await contract_creation_block_async(self.address):
            logger.debug("%s was not deployed at block %s", self, block)
            return 0
        index = await self.get_coin_index(token, sync=False)
        balance = await self._get_balance(index, block) or 0
        logger.debug("%s liquidity for %s at block %s is %s", self, token, block, balance)
        return balance


class CurveRegistry(a_sync.ASyncGenericSingleton):
    __slots__ = ("__task",)

    def __init__(self, *, asynchronous: bool = False) -> None:
        super().__init__()
        self.asynchronous = asynchronous
        try:
            self.address_provider = AddressProvider(
                ADDRESS_PROVIDER, asynchronous=self.asynchronous
            )
            self.address_provider.contract
        except (ContractNotFound, ContractNotVerified) as e:
            raise UnsupportedNetwork("curve is not supported on this network") from e
        except MessedUpBrownieContract as e:
            if CHAINID == Network.Cronos:
                raise UnsupportedNetwork("curve is not supported on this network") from e
            else:
                raise

        self.registries = defaultdict(set)  # registry -> pools
        self.factories = defaultdict(set)  # factory -> pools
        self.pools = set()
        self.token_to_pool = dict()  # lp_token -> pool

    def __repr__(self) -> str:
        return "<CurveRegistry>"

    @property
    def identifiers(self) -> List[EthAddress]:
        return self.address_provider.identifiers

    @a_sync.aka.cached_property
    async def registry(self) -> Contract:
        try:
            return await Contract.coroutine(self.identifiers[0][-1])
        except IndexError:  # if we couldn't get the registry via logs
            return await Contract.coroutine(
                await raw_call(
                    self.address_provider,
                    "get_registry()",
                    output="address",
                    sync=False,
                )
            )

    __registry__: HiddenMethodDescriptor[Self, Contract]

    async def load_all(self) -> None:
        await self._done.wait()

    async def get_factory(self, pool: AddressOrContract) -> Contract:
        """
        Get metapool factory that has spawned a pool.

        Args:
            pool: The address or contract of the pool.

        Returns:
            The :class:`~y.contracts.Contract` representing the factory.

        Examples:
            >>> factory = await curve.get_factory("0x1234567890abcdef1234567890abcdef12345678")
            >>> print(factory)
            <Contract '0x...'>
        """
        try:
            factory = next(
                factory
                for factory, factory_pools in self.factories.items()
                if str(pool) in factory_pools
            )
            return await Contract.coroutine(factory)
        except StopIteration:
            return None

    @a_sync_ttl_cache
    async def get_price(
        self,
        token: Address,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[float]:
        pool: CurvePool = await self.get_pool(token, sync=False)
        if pool is None:
            return None
        tvl = await pool.get_tvl(block=block, skip_cache=skip_cache, sync=False)
        if tvl is None:
            return None
        try:
            return tvl / await ERC20(token, asynchronous=True).total_supply_readable(block)
        except ZeroDivisionError:
            return None

    @a_sync.a_sync(cache_type="memory")
    async def get_pool(self, token: AnyAddressType) -> CurvePool:
        """
        Get Curve pool (swap) address by LP token address. Supports factory pools.

        Args:
            token: The address of the LP token.

        Returns:
            The :class:`~y.prices.stable_swap.curve.CurvePool` associated with the LP token.

        Examples:
            >>> pool = await curve.get_pool("0x1234567890abcdef1234567890abcdef12345678")
            >>> print(pool)
            <CurvePool '0x...'>
        """
        await self.load_all()

        token = await convert.to_address_async(token)
        if token in self.token_to_pool and token != ZERO_ADDRESS:
            return CurvePool(self.token_to_pool[token], asynchronous=self.asynchronous)

    @a_sync.a_sync(cache_type="memory")
    async def get_price_for_underlying(
        self,
        token_in: Address,
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        try:
            pools = (await self.__coin_to_pools__)[token_in]
        except KeyError:
            return None

        for pool in ignore_pools:
            try:
                pools.remove(pool)
            except ValueError:
                continue

        if pools and block is not None:
            pools = [
                pool
                async for pool, deploy_block in CurvePool.deploy_block.map(
                    pools, when_no_history_return_0=True
                )
                if deploy_block <= block
            ]

        if not pools:
            return None
        # Choose a pool to use for pricing `token_in`.
        elif len(pools) == 1:
            pool = pools[0]
        else:
            # Use the pool with deepest liquidity.
            deepest_pool, deepest_bal = None, 0
            async for pool, depth in CurvePool.check_liquidity.map(
                pools, token=token_in, block=block
            ).map():
                if depth > deepest_bal:
                    deepest_pool = pool
                    deepest_bal = depth
            pool = deepest_pool

        if pool is None:
            return None

        if len(await pool.__coins__) != 2:
            # TODO: handle this sitch if necessary
            return

        # Get the price for `token_in` using the selected pool.
        # this works for most typical metapools

        token_in_ix = await pool.get_coin_index(token_in, sync=False)
        token_out_ix = 0 if token_in_ix == 1 else 1 if token_in_ix == 0 else None
        dy: Optional[WeiBalance] = await pool.get_dy(
            token_in_ix,
            token_out_ix,
            block=block,
            ignore_pools=ignore_pools,
            skip_cache=skip_cache,
            sync=False,
        )
        if dy is None:
            return None

        try:
            return await dy.__value_usd__
        except yPriceMagicError as e:
            logger.debug("%s for %s at block %s", type(e.exception).__name__, token_in, block)
            if not isinstance(e.exception, PriceError):
                raise

            # try to get price from a different pool
            return await self.get_price_for_underlying(
                token_in,
                block,
                ignore_pools=(*ignore_pools, pool),
                skip_cache=skip_cache,
            )

    @a_sync.aka.cached_property
    async def coin_to_pools(self) -> Dict[str, List[CurvePool]]:
        mapping = defaultdict(set)
        await self.load_all()
        for pool in {CurvePool(pool) for pools in self.factories.values() for pool in pools}:
            for coin in await pool.__coins__:
                mapping[coin].add(pool)
        return {coin: list(pools) for coin, pools in mapping.items()}

    __coin_to_pools__: HiddenMethodDescriptor[Self, Dict[str, List[CurvePool]]]

    async def check_liquidity(
        self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...]
    ) -> int:
        if pools_for_token := (await self.__coin_to_pools__).get(token):
            if pools := list(filterfalse(ignore_pools.__contains__, pools_for_token)):
                return await CurvePool.check_liquidity.max(
                    pools, token=token, block=block, sync=False
                )
        return 0

    @cached_property
    def _done(self) -> a_sync.Event:
        """A helper function to ensure the Event is attached to the correct loop."""
        self._task
        return a_sync.Event(name="curve")

    @cached_property
    def _task(self) -> Task:
        _startup_logger_debug("creating loader task for %s", self)
        task = create_task(coro=self._load_all(), name=f"{self}._load_all()")

        def done_callback(t: Task):
            if e := t.exception():
                startup_logger.error("exception while loading %s: %s", self, e)
                startup_logger.exception(e)
                self.__task = None
                raise e

        task.add_done_callback(done_callback)
        return task

    async def _load_all(self) -> None:
        await self.address_provider
        _startup_logger_debug(
            "curve address provider events loaded, now loading factories and pools"
        )
        # NOTE: Gnosis chain's address provider fails to provide registry via events. Maybe other chains as well.
        if (
            not self.identifiers[Ids.Main_Registry]
            and (registry := await self.address_provider.get_registry()) != ZERO_ADDRESS
        ):
            self.identifiers[Ids.Main_Registry] = [registry]
        while True:
            # Check if any registries were updated, then ensure all old and new are loaded
            if registries := [
                self.identifiers[i][-1]
                for i in [Ids.Main_Registry, Ids.CryptoSwap_Registry]
                if self.identifiers[i]
            ]:
                await a_sync.map(Registry, registries, curve=self, asynchronous=self.asynchronous)
            # load metapool and curve v5 factories
            await self.address_provider._load_factories()
            await sleep(600)


try:
    curve: CurveRegistry = CurveRegistry(asynchronous=True)
except UnsupportedNetwork:
    curve = set()


__startup_logger_log = startup_logger._log


def _startup_logger_log_debug(msg: str, *args: Any) -> None:
    __startup_logger_log(DEBUG, msg, args)



"""
Utility functions for retrieving Ethereum client information.
"""

import logging

from async_lru import alru_cache
from brownie import web3
from cachetools.func import lru_cache
from dank_mids import dank_web3
from web3._utils.rpc_abi import RPC

from y.utils.logging import yLazyLogger

logger = logging.getLogger(__name__)


@lru_cache(1)
@yLazyLogger(logger)
def get_ethereum_client() -> str:
    """
    Get the Ethereum client type for the current connection.

    Returns:
        A string representing the Ethereum client type, such as 'geth', 'erigon', or 'tg'.

    Examples:
        >>> from y.utils.client import get_ethereum_client
        >>> client = get_ethereum_client()
        >>> print(client)
        'geth'

        >>> # Example with a different client
        >>> client = get_ethereum_client()
        >>> print(client)
        'erigon'

    See Also:
        - :func:`get_ethereum_client_async` for the asynchronous version of this function.
    """
    try:
        # web3py < 6.0
        version = web3.clientVersion
    except AttributeError:
        # web3py >= 6.0
        version = web3.client_version
    return _get_ethereum_client(version)


@alru_cache(maxsize=1)
async def get_ethereum_client_async() -> str:
    """
    Asynchronously get the Ethereum client type for the current connection.

    Returns:
        A string representing the Ethereum client type, such as 'geth', 'erigon', or 'tg'.

    Examples:
        >>> from y.utils.client import get_ethereum_client_async
        >>> client = await get_ethereum_client_async()
        >>> print(client)
        'erigon'

        >>> # Example with a different client
        >>> client = await get_ethereum_client_async()
        >>> print(client)
        'tg'

    See Also:
        - :func:`get_ethereum_client` for the synchronous version of this function.
    """
    return _get_ethereum_client(await dank_web3.manager.coro_request(RPC.web3_clientVersion, []))


@yLazyLogger(logger)
def _get_ethereum_client(client: str) -> str:
    """
    Determine the Ethereum client type based on the client version string.

    Args:
        client: A string containing the client version information.

    Returns:
        A string representing the Ethereum client type, such as 'geth', 'erigon', or 'tg'.

    Examples:
        >>> from y.utils.client import _get_ethereum_client
        >>> client_type = _get_ethereum_client("Geth/v1.10.23-stable/linux-amd64/go1.18.7")
        >>> print(client_type)
        'geth'

        >>> # Example with a different client version
        >>> client_type = _get_ethereum_client("Erigon/v2021.07.2/linux-amd64/go1.16.5")
        >>> print(client_type)
        'erigon'
    """
    logger.debug("client: %s", client)
    return next(
        (identifier for prefix, identifier in _clients.items() if client.startswith(prefix)),
        client,
    )


_clients = {
    "geth": "geth",
    "erigon": "erigon",
    "TurboGeth": "tg",
}



import logging
from datetime import datetime
from decimal import Decimal, InvalidOperation
from typing import TYPE_CHECKING, Any, Callable, TypeVar
from typing import Optional as typing_Optional

from pony.orm import (
    Database,
    InterfaceError,
    Optional,
    PrimaryKey,
    Required,
    Set,
    TransactionIntegrityError,
    commit,
    composite_index,
    db_session,
)
from typing_extensions import ParamSpec

from y._db.decorators import retry_locked, ydb_write_threads

db = Database()

# makes type checking work, see below for info:
# https://pypi.org/project/pony-stubs/
DbEntity = db.Entity


logger = logging.getLogger(__name__)


class _AsyncEntityMixin:
    """Mixin class intended for future asynchronous operations.

    This class is currently not implemented. It is designed to provide
    asynchronous capabilities to database entities in the future.

    TODO: Implement asynchronous methods for database operations.

    Examples:
        Intended usage once implemented:
        >>> class MyEntity(DbEntity, _AsyncEntityMixin):
        ...     pass

    See Also:
        - :mod:`a_sync` for asynchronous programming utilities.

    @classmethod
    @a_sync('async')
    def get(cls: E, *args, **kwargs) -> E:
        return super(DbEntity).get(*args, **kwargs)
    @classmethod
    @a_sync('async')
    def select(cls: E, *args, **kwargs) -> E:
        return super(DbEntity).select(*args, **kwargs)
    """


class Chain(DbEntity, _AsyncEntityMixin):
    """Represents a blockchain chain entity in the database."""

    id = PrimaryKey(int)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        blocks: Set["Block"]
        addresses: Set["Address"]
        log_cached: Set["LogCacheInfo"]
        trace_caches: Set["TraceCacheInfo"]

    blocks = Set("Block")
    addresses = Set("Address")
    log_cached = Set("LogCacheInfo")
    trace_caches = Set("TraceCacheInfo")


class Block(DbEntity, _AsyncEntityMixin):
    """Represents a block entity in the database."""

    chain = Required(Chain, reverse="blocks")
    number = Required(int)
    PrimaryKey(chain, number)
    hash = Optional(int, lazy=True)
    timestamp = Optional(datetime, lazy=True)

    composite_index(chain, hash)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        prices: Set["Price"]
        contracts_deployed: Set["Contract"]
        logs: Set["Log"]
        traces: Set["Trace"]

    prices = Set("Price", reverse="block", cascade_delete=False)
    contracts_deployed = Set("Contract", reverse="deploy_block")
    logs = Set("Log", reverse="block", cascade_delete=False)
    traces = Set("Trace", reverse="block", cascade_delete=False)


class Address(DbEntity, _AsyncEntityMixin):
    """Represents an address entity in the database."""

    chain = Required(Chain, lazy=True, reverse="addresses")
    address = Required(str, 42, lazy=True)
    PrimaryKey(chain, address)
    notes = Optional(str, lazy=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        contracts_deployed: Set["Contract"]

    contracts_deployed = Set("Contract", reverse="deployer")


class Contract(Address):
    """Represents a contract entity in the database."""

    deployer = Optional(Address, reverse="contracts_deployed", lazy=True, cascade_delete=False)
    deploy_block = Optional(Block, reverse="contracts_deployed", lazy=True, cascade_delete=False)


class Token(Contract):
    """Represents a token entity in the database."""

    symbol = Optional(str, lazy=True)
    name = Optional(str, lazy=True)
    decimals = Optional(int, lazy=True)
    bucket = Optional(str, index=True, lazy=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        prices: Set["Price"]

    prices = Set("Price", reverse="token")


class Price(DbEntity):
    """Represents a price entity in the database."""

    block = Required(Block, index=True, lazy=True)
    token = Required(Token, index=True, lazy=True)
    PrimaryKey(block, token)
    price = Required(Decimal, 38, 18)


class TraceCacheInfo(DbEntity):
    """Represents trace cache information in the database."""

    chain = Required(Chain, index=True)
    to_addresses = Required(bytes, index=True)
    from_addresses = Required(bytes, index=True)
    PrimaryKey(chain, to_addresses, from_addresses)
    cached_from = Required(int)
    cached_thru = Required(int)


class LogCacheInfo(DbEntity):
    """Represents log cache information in the database."""

    chain = Required(Chain, index=True)
    address = Required(str, 42, index=True)
    topics = Required(bytes)
    PrimaryKey(chain, address, topics)
    cached_from = Required(int)
    cached_thru = Required(int)


class LogTopic(DbEntity):
    """Represents a log topic entity in the database.

    Just makes the :ref:`Log` db smaller.
    """

    dbid = PrimaryKey(int, size=64, auto=True)
    topic = Required(str, 64, unique=True, lazy=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        logs_as_topic0: Set["Log"]
        logs_as_topic1: Set["Log"]
        logs_as_topic2: Set["Log"]
        logs_as_topic3: Set["Log"]

    logs_as_topic0 = Set("Log", reverse="topic0")
    logs_as_topic1 = Set("Log", reverse="topic1")
    logs_as_topic2 = Set("Log", reverse="topic2")
    logs_as_topic3 = Set("Log", reverse="topic3")


class Hashes(DbEntity):
    """Represents a hash entity in the database.

    Just makes :ref:`Log` pk and indexes smaller.
    """

    dbid = PrimaryKey(int, size=64, auto=True)
    hash = Required(str, 64, unique=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        logs_for_tx: Set["Log"]
        logs_for_addr: Set["Log"]

    logs_for_tx = Set("Log", reverse="tx")
    logs_for_addr = Set("Log", reverse="address")


class Log(DbEntity):
    """Represents a log entity in the database."""

    block = Required(Block, index=True, lazy=True)
    tx = Required(Hashes, lazy=True)
    log_index = Required(int, size=16, lazy=True)
    PrimaryKey(block, tx, log_index)

    address = Required(Hashes, index=True, lazy=True)

    topic0 = Required(LogTopic, index=True, lazy=True)
    topic1 = Optional(LogTopic, index=True, lazy=True)
    topic2 = Optional(LogTopic, index=True, lazy=True)
    topic3 = Optional(LogTopic, index=True, lazy=True)
    composite_index(address, topic0)
    composite_index(topic0, topic1)
    composite_index(topic0, topic2)
    composite_index(topic0, topic3)
    composite_index(block, address, topic0)
    composite_index(block, topic0, topic1)
    composite_index(block, topic0, topic2)
    composite_index(block, topic0, topic3)
    composite_index(topic0, topic1, topic2, topic3)
    composite_index(block, topic0, topic1, topic2, topic3)
    raw = Required(bytes, lazy=True)


class Trace(DbEntity):
    """Represents a trace entity in the database."""

    id = PrimaryKey(int, auto=True)
    block = Required(Block, index=True, lazy=True)
    hash = Required(str, 64, index=True, lazy=True)
    from_address = Required(str, 42, index=True, lazy=True)
    to_address = Required(str, 42, index=True, lazy=True)
    raw = Required(bytes)


class BlockAtTimestamp(DbEntity):
    """Represents a block at a specific timestamp in the database."""

    chainid = Required(int)
    timestamp = Required(datetime)
    PrimaryKey(chainid, timestamp)
    block = Required(int)


__write_threads_submit = ydb_write_threads.submit


def insert_nowait(type: DbEntity, **kwargs: Any) -> typing_Optional[DbEntity]:
    # sourcery skip: simplify-boolean-comparison
    __write_threads_submit(exc_wrapper, insert, type, **kwargs, fire_and_forget=True)


__P = ParamSpec("__P")
__T = TypeVar("__T")


def exc_wrapper(
    func: Callable[__P, __T], *args: __P.args, **kwargs: __P.kwargs
) -> typing_Optional[DbEntity]:
    try:
        return func(*args, **kwargs)
    except InvalidOperation:
        pass
    except Exception as e:
        logger.exception(e)


@db_session
@retry_locked
def insert(type: DbEntity, **kwargs: Any) -> typing_Optional[DbEntity]:
    """Inserts a new entity into the database with retry logic.

    This function attempts to insert a new entity of the specified type into the database.
    It includes retry logic to handle specific database exceptions.

    Args:
        type: The type of the entity to insert.
        **kwargs: The attributes of the entity to insert.

    Examples:
        >>> new_block = insert(Block, chain=chain_instance, number=123456)
        >>> new_token = insert(Token, symbol="ETH", name="Ethereum")

    See Also:
        - :func:`pony.orm.commit`
        - :func:`pony.orm.db_session`
        - :func:`retry_locked`
    """
    try:
        while True:
            try:
                entity = type(**kwargs)
                commit()
                logger.debug("inserted %s to db", entity)
                return entity
            except InterfaceError as e:
                logger.debug("%s while inserting %s", e, type.__name__)
            except InvalidOperation as e:
                e.args = *e.args, type, kwargs
                raise
    except TransactionIntegrityError as e:
        constraint_errs = (
            "UNIQUE constraint failed",
            "duplicate key value violates unique constraint",
        )
        if any(map((msg := str(e)).__contains__, constraint_errs)):
            logger.debug("%s: %s %s", msg, type.__name__, kwargs)
        else:
            logger.debug(
                "%s %s when inserting %s",
                e.__class__.__name__,
                str(e),
                e,
                type.__name__,
            )
            raise



import abc
from decimal import Decimal
from typing import Optional

from a_sync import cgather

import y.ENVIRONMENT_VARIABLES as ENVS
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20
from y.datatypes import Block, UsdPrice, UsdValue


class Wrapper(ERC20):
    """
    An abstract base class representing an ERC20 token that holds balances of (an)other ERC20 token(s).

    This class extends the :class:`~y.classes.common.ERC20` class to represent wrapper tokens, which are tokens that derive their value
    from holding other tokens. The specific implementation details are left to the subclasses.

    Note:
        This class is currently a placeholder and does not implement any additional methods.

    See Also:
        - :class:`~y.classes.common.ERC20`
    """

    # TODO: implement


class LiquidityPool(Wrapper):
    """
    An abstract base class representing a liquidity pool that pools multiple ERC20s together for swapping.

    This class extends the :class:`~Wrapper` class to represent liquidity pools, which are special types of wrapper
    tokens that allow for token swaps. It provides methods for getting the total value locked (TVL) and pool price.

    Examples:
        >>> class MyLiquidityPool(LiquidityPool):
        ...     async def get_tvl(self, block=None, skip_cache=False):
        ...         return UsdValue(1000000)
        ...
        >>> pool = MyLiquidityPool("0xAddress")
        >>> price = await pool.get_pool_price()
        >>> print(price)
        1000000.0

    See Also:
        - :class:`~Wrapper`
    """

    # TODO: implement this elsewhere outside of just balancer

    @stuck_coro_debugger
    async def get_pool_price(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Optional[UsdPrice]:
        """
        Calculate the price of the liquidity pool token.

        This method calculates the price of a single liquidity pool token by dividing the total value locked (TVL)
        by the total supply of the pool tokens.

        Args:
            block: The block number at which to calculate the price. If None, uses the latest block.
            skip_cache: If True, bypasses ypricemagic's local caching mechanisms and forces a fresh calculation.

        Returns:
            The price of a single liquidity pool token as a :class:`~y.datatypes.UsdPrice` object.

        Examples:
            >>> class MyLiquidityPool(LiquidityPool):
            ...     async def get_tvl(self, block=None, skip_cache=False):
            ...         return UsdValue(1000000)
            ...
            >>> pool = MyLiquidityPool("0xAddress")
            >>> price = await pool.get_pool_price()
            >>> print(price)
            1000000.0

        See Also:
            - :meth:`get_tvl`
        """
        tvl, total_supply = await cgather(
            self.get_tvl(block=block, skip_cache=skip_cache, sync=False),
            self.total_supply_readable(block=block, sync=False),
        )
        return None if tvl is None else UsdPrice(Decimal(tvl) / Decimal(total_supply))

    @abc.abstractmethod
    async def get_tvl(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Optional[UsdValue]:
        """
        Get the Total Value Locked (TVL) in the liquidity pool.

        This is an abstract method that must be implemented by subclasses. It should return the total value
        of all assets locked in the liquidity pool.

        Args:
            block: The block number at which to calculate the TVL. If None, uses the latest block.
            skip_cache: If True, bypasses ypricemagic's local caching mechanisms and forces a fresh calculation.

        Returns:
            The Total Value Locked (TVL) in the pool as a :class:`~y.datatypes.UsdValue` object.

        Note:
            The specific implementation of this method will depend on the type of liquidity pool.

        Examples:
            >>> class MyLiquidityPool(LiquidityPool):
            ...     async def get_tvl(self, block=None, skip_cache=False):
            ...         return UsdValue(1000000)
            ...
            >>> pool = MyLiquidityPool("0xAddress")
            >>> tvl = await pool.get_tvl()
            >>> print(tvl)
            1000000.0
        """



from asyncio import Lock
from hashlib import sha1
from json import dumps, loads
from pathlib import Path
from typing import Any, Container, Dict, Literal, Optional, Tuple

import aiosqlite
from a_sync import SmartProcessingQueue
from a_sync.async_property.cached import async_cached_property
from brownie._config import CONFIG, _get_data_folder
from brownie.exceptions import BrownieEnvironmentError
from brownie.network.contract import _resolve_address
from brownie.project.build import DEPLOYMENT_KEYS
from eth_utils.toolz import keymap
from functools import lru_cache
from sqlite3 import InterfaceError, OperationalError

from y.datatypes import Address


SourceKey = Literal[
    "address",
    "alias",
    "paths",
    "abi",
    "ast",
    "bytecode",
    "compiler",
    "contractName",
    "deployedBytecode",
    "deployedSourceMap",
    "language",
    "natspec",
    "opcodes",
    "pcMap",
    "sourceMap",
    "type",
]

BuildJson = Dict[str, Any]
Sources = Dict[SourceKey, Any]


SOURCE_KEYS: Tuple[SourceKey, ...] = SourceKey.__args__

DISCARD_SOURCE_KEYS: Tuple[SourceKey, ...] = (
    "ast",
    "bytecode",
    "coverageMap",
    "deployedBytecode",
    "deployedSourceMap",
    "natspec",
    "opcodes",
    "pcMap",
)
"""
To keep our Contract objects smaller, we do not load all contract source data. Many are not relevant for our data analysis type use case.

These keys will not be included in y.Contract object build data. If you need them, consider just using a dank_mids.Contract for your use case.
"""


sqlite_lock = Lock()


class AsyncCursor:
    def __init__(self, filename):
        self._filename = filename

    @async_cached_property
    async def connect(self):
        """Establish an async connection to the SQLite database"""
        async with sqlite_lock:
            self._db = await aiosqlite.connect(self._filename, isolation_level=None)

    async def insert(self, table, *values):
        raise NotImplementedError
        await self.connect

        # Convert any dictionaries/lists to JSON strings before inserting
        values = [dumps(val) if isinstance(val, (dict, list)) else val for val in values]

        # Prepare the parameter placeholders
        placeholders = ",".join("?" * len(values))
        query = f"INSERT OR REPLACE INTO {table} VALUES ({placeholders})"

        # Execute the query and commit the changes
        async with self._db.execute(query, values):
            await self._db.commit()

    async def fetchone(self, cmd: str, *args) -> Optional[Tuple]:
        await self.connect
        async with sqlite_lock:
            async with self._db.execute(cmd, args) as cursor:
                if row := await cursor.fetchone():
                    # Convert any JSON-serialized columns back to their original data structures
                    return tuple(loads(i) if str(i).startswith(("[", "{")) else i for i in row)


cur = AsyncCursor(_get_data_folder().joinpath("deployments.db"))
fetchone = SmartProcessingQueue(cur.fetchone, num_workers=32)


@lru_cache(maxsize=None)
def _get_select_statement() -> str:
    try:
        return f"SELECT * FROM chain{CONFIG.active_network['chainid']}"
    except KeyError:
        raise BrownieEnvironmentError("Functionality not available in local environment") from None


async def _get_deployment(
    address: str = None,
    alias: str = None,
    skip_source_keys: Container[SourceKey] = DISCARD_SOURCE_KEYS,
) -> Tuple[Optional[BuildJson], Optional[Sources]]:
    if address and alias:
        raise ValueError("Passed both params address and alias, should be only one!")
    if address:
        address = _resolve_address(address)
        where_clause = f" WHERE address='{address}'"
    elif alias:
        where_clause = f" WHERE alias='{alias}'"

    try:
        row = await fetchone(_get_select_statement() + where_clause)
    except OperationalError:
        row = None
    if not row:
        return None, None

    build_json = dict(zip(SOURCE_KEYS, row))
    path_map = build_json.pop("paths")

    sources = {
        source_key: await __fetch_source_for_hash(val)
        for val, source_key in path_map.values()
        if source_key not in skip_source_keys
    }

    build_json["allSourcePaths"] = {k: v[1] for k, v in path_map.items()}
    if isinstance(build_json.get("pcMap"), dict):
        build_json["pcMap"] = keymap(int, build_json["pcMap"])

    return build_json, sources


async def __fetch_source_for_hash(hashval: str) -> Any:
    return (await fetchone("SELECT source FROM sources WHERE hash=?", hashval))[0]



"""Tests for Uniswap price fetching functions.

This module contains asynchronous tests for verifying price retrieval via the Uniswap multiplexer
across different Uniswap versions (V1, V2, and V3). In particular, note that for Uniswap V2 the test now uses
the primary router determined via liquidity depth via :meth:`~y.prices.dex.uniswap.uniswap.UniswapMultiplexer.routers_by_depth`.

See Also:
    :class:`~y.prices.dex.uniswap.uniswap.UniswapMultiplexer`
"""

import pytest
from a_sync import cgather
from brownie import chain

from tests.fixtures import async_uni_v1, mutate_addresses
from y.networks import Network
from y.prices import magic
from y.prices.dex.uniswap import v3
from y.prices.dex.uniswap.uniswap import uniswap_multiplexer

V1_TOKENS = {
    Network.Mainnet: [
        "0x6B175474E89094C44Da98b954EedeAC495271d0F",
        "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
    ],
}.get(chain.id, [])

V2_TOKENS = {
    Network.Mainnet: [
        "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",
        "0xBA11D00c5f74255f56a5E366F4F77f5A186d7f55",
        "0xc00e94Cb662C3520282E6f5717214004A7f26888",
        "0xD533a949740bb3306d119CC777fa900bA034cd52",
        "0x6B175474E89094C44Da98b954EedeAC495271d0F",
        "0x6810e776880C02933D47DB1b9fc05908e5386b96",
        "0xc944E90C64B2c07662A292be6244BDf05Cda44a7",
        "0x514910771AF9Ca656af840dff83E8264EcF986CA",
        "0x0F5D2fB29fb7d3CFeE444a200298f468908cC942",
        "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
        "0xec67005c4E498Ec7f55E092bd1d35cbC47C91892",
        "0x4fE83213D56308330EC302a8BD641f1d0113A4Cc",
        "0xC011a73ee8576Fb46F5E1c5751cA3B9Fe0af2a6F",
        "0x04Fa0d235C4abf4BcF4787aF4CF447DE572eF828",
        "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
        "0xdAC17F958D2ee523a2206206994597C13D831ec7",
        "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
        "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
        "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e",
        "0xE41d2489571d322189246DaFA5ebDe1F4699F498",
    ],
}.get(chain.id, [])

V2_TOKENS = mutate_addresses(V2_TOKENS)


@pytest.mark.parametrize("token", V1_TOKENS)
@pytest.mark.asyncio_cooperative
async def test_uniswap_v1(token, async_uni_v1):
    """Test Uniswap V1 price fetching.

    This test concurrently retrieves the price using an asynchronous Uniswap V1 caller and the generic
    :func:`~y.prices.magic.get_price` function. It verifies that the prices obtained from both methods are
    consistent within a 5% relative tolerance.

    Args:
        token: The token address to query.
        async_uni_v1: Fixture providing an asynchronous Uniswap V1 instance.

    See Also:
        :meth:`~y.prices.dex.uniswap.uniswap.UniswapMultiplexer.get_price`
    """
    price, alt_price = await cgather(
        async_uni_v1.get_price(token, None),
        magic.get_price(token, skip_cache=True, sync=False),
    )
    print(token, price, alt_price)
    # check if price is within 5% range
    assert price == pytest.approx(alt_price, rel=5e-2)


@pytest.mark.parametrize("token", V2_TOKENS)
@pytest.mark.asyncio_cooperative
async def test_uniswap_v2(token):
    """Test Uniswap V2 price fetching.

    This test retrieves the list of routers sorted by liquidity depth for the provided token by calling
    :meth:`~y.prices.dex.uniswap.uniswap.UniswapMultiplexer.routers_by_depth`, and then concurrently fetches
    price data from the selected router and from the generic :func:`~y.prices.magic.get_price` function.
    The test verifies that the prices obtained are consistent within a 5% relative tolerance.

    Args:
        token: The token address to query.

    See Also:
        :meth:`~y.prices.dex.uniswap.uniswap.UniswapMultiplexer.routers_by_depth`
    """
    deepest_router = await uniswap_multiplexer.deepest_router(token, sync=False)
    price, alt_price = await cgather(
        deepest_router.get_price(token, skip_cache=True, sync=False),
        magic.get_price(token, skip_cache=True, sync=False),
    )
    print(token, price, alt_price)
    assert price == pytest.approx(alt_price, rel=5e-2)


@pytest.mark.parametrize("token", V2_TOKENS)
@pytest.mark.asyncio_cooperative
async def test_uniswap_v3(token):
    """Test Uniswap V3 price fetching.

    This test concurrently retrieves pricing data using the Uniswap V3 router and the generic
    :func:`~y.prices.magic.get_price` function. It ensures that the price returned by the Uniswap V3 router
    is within a 5% relative tolerance of the price obtained from the magic fetch.

    Args:
        token: The token address to query.

    See Also:
        :meth:`~y.prices.dex.uniswap.v3.uniswap_v3.get_price`
    """
    price, alt_price = await cgather(
        v3.uniswap_v3.get_price(token, skip_cache=True, sync=False),
        magic.get_price(token, skip_cache=True, sync=False),
    )
    print(token, price, alt_price)
    assert price == pytest.approx(alt_price, rel=5e-2)



import logging
import time
from functools import lru_cache, wraps
from typing import Callable, Iterable, TypeVar
from typing_extensions import ParamSpec

from a_sync import PruningThreadPoolExecutor, a_sync
from a_sync.a_sync import ASyncFunction
from brownie import chain
from pony.orm import (
    CommitException,
    OperationalError,
    TransactionError,
    UnexpectedError,
    commit,
    db_session,
)


_T = TypeVar("_T")
_P = ParamSpec("_P")

logger = logging.getLogger(__name__)

ydb_read_threads = PruningThreadPoolExecutor(12)
ydb_write_threads = PruningThreadPoolExecutor(12)


def retry_locked(callable: Callable[_P, _T]) -> Callable[_P, _T]:
    """Retries a database operation if it encounters specific exceptions related to database locks.

    This decorator wraps a function and retries it with an increasing sleep interval if it encounters
    `CommitException`, `OperationalError`, or `UnexpectedError` with a "database is locked" message.
    It also handles `TransactionError` by checking for specific error messages related to transaction mixing.

    Args:
        callable: The function to be wrapped and retried.

    Returns:
        The wrapped function with retry logic.

    Examples:
        >>> @retry_locked
        ... def my_db_function():
        ...     # perform some database operations
        ...     pass

    See Also:
        - :func:`pony.orm.commit`
        - :func:`pony.orm.db_session`
    """

    @wraps(callable)
    def retry_locked_wrap(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        sleep = 0.05
        while True:
            try:
                retval = callable(*args, **kwargs)
                commit()
                return retval
            except (CommitException, OperationalError, UnexpectedError) as e:
                log = logger.warning if sleep > 1 else logger.debug
                log("%s.%s got exc %s", callable.__module__, callable.__name__, e)
                if "database is locked" not in str(e):
                    raise
                time.sleep(sleep)
                sleep *= 1.5
            except TransactionError as e:
                logger.debug("%s.%s got exc %s", callable.__module__, callable.__name__, e)
                if "An attempt to mix objects belonging to different transactions" not in str(e):
                    raise

    return retry_locked_wrap


db_session_retry_locked = lambda func: retry_locked(db_session(retry_locked(func)))

a_sync_read_db_session: Callable[[Callable[_P, _T]], ASyncFunction[_P, _T]] = lambda fn: a_sync(
    default="async", executor=ydb_read_threads
)(db_session_retry_locked(fn))
"""Decorator for asynchronous read database sessions with retry logic.

This decorator wraps a function with an asynchronous database session for read operations,
applying retry logic for handling database locks.

Args:
    fn: The function to be wrapped.

Returns:
    An asynchronous function with read database session management and retry logic.

Examples:
    >>> @a_sync_read_db_session
    ... async def read_data():
    ...     # perform read operations
    ...     pass

See Also:
    - :func:`retry_locked`
    - :func:`pony.orm.db_session`
"""


db_session_cached = lambda func: retry_locked(
    lru_cache(maxsize=None)(db_session(retry_locked(func)))
)


_result_count_logger = logging.getLogger(f"{__name__}.result_count")
_CHAIN_INFO = "chain", chain.id


def log_result_count(
    name: str, arg_names: Iterable[str] = []
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    """Logs the number of results returned by a function.

    This decorator logs the number of results returned by a function at the DEBUG level,
    including additional context such as the chain ID and argument values.

    Args:
        name: The name of the result being logged.
        arg_names: An iterable of argument names to include in the log.

    Returns:
        A decorator that logs the result count of the wrapped function.

    Examples:
        >>> @log_result_count("items", ["arg1", "arg2"])
        ... def get_items(arg1, arg2):
        ...     return [1, 2, 3]

    See Also:
        - :mod:`logging`
    """

    def result_count_deco(fn: Callable[_P, _T]) -> Callable[_P, _T]:
        @wraps(fn)
        def result_count_wrap(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            results = fn(*args, **kwargs)
            if _result_count_logger.isEnabledFor(logging.DEBUG):
                arg_values = " ".join(f"{k} {v}" for k, v in (_CHAIN_INFO, *zip(arg_names, args)))
                _result_count_logger.debug("loaded %s %s for %s", len(results), name, arg_values)
            return results

        return result_count_wrap

    return result_count_deco



import time

import pytest
from brownie import ZERO_ADDRESS, chain

from tests.fixtures import mainnet_only
from y.contracts import contract_creation_block_async
from y.networks import Network
from y.prices.chainlink import FEEDS, chainlink

feeds = set(FEEDS.keys())
feeds.update(
    {
        # Add feeds that *should* come from registry
        Network.Mainnet: [
            "0x0000000000000000000000000000000000000024",
            "0x000000000000000000000000000000000000007c",
            "0x000000000000000000000000000000000000009c",
            "0x0000000000000000000000000000000000000164",
            "0x0000000000000000000000000000000000000188",
            "0x000000000000000000000000000000000000019a",
            "0x000000000000000000000000000000000000022A",
            "0x0000000000000000000000000000000000000236",
            "0x0000000000000000000000000000000000000260",
            # "0x0000000000000000000000000000000000000283", chainlink feed disabled at block 14673916 - 0xd1898776f4f2333c26c170b58ec98430d29c224a81256fd70add5db05a036b20
            "0x00000000000000000000000000000000000002be",
            "0x00000000000000000000000000000000000002c6",
            "0x00000000000000000000000000000000000002F4",
            "0x000000000000000000000000000000000000033a",
            "0x00000000000000000000000000000000000003d2",
            "0x00000000000000000000000000000000000003Da",
            "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e",
            "0x0D8775F648430679A709E98d2b0Cb6250d2887EF",
            "0x111111111117dC0aa78b770fA6A738034120C302",
            "0x1776e1F26f98b1A5dF9cD347953a26dd3Cb46671",
            "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C",
            "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
            "0x221657776846890989a759BA2973e427DfF5C9bB",
            "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
            "0x3155BA85D5F96b2d030a4966AF206230e46849cb",
            "0x33D0568941C0C64ff7e0FB4fbA0B11BD37deEd9f",
            "0x3432B6A60D23Ca0dFCa7761B7ab56459D9C964D0",
            "0x3472A5A71965499acd81997a54BBA8D852C6E53d",
            "0x3845badAde8e6dFF049820680d1F14bD3903a5d0",
            "0x408e41876cCCDC0F92210600ef50372656052a38",
            "0x4575f41308EC1483f3d399aa9a2826d74Da13Deb",
            "0x4688a8b1F292FDaB17E9a90c8Bc379dC1DBd8713",
            "0x4C19596f5aAfF459fA38B0f7eD92F11AE6543784",
            "0x4Fabb145d64652a948d72533023f6E7A623C7C53",
            "0x514910771AF9Ca656af840dff83E8264EcF986CA",
            "0x57Ab1ec28D129707052df4dF418D58a2D46d5f51",
            "0x584bC13c7D411c00c01A62e8019472dE68768430",
            "0x674C6Ad92Fd080e4004b2312b45f796a192D27a0",
            "0x6B175474E89094C44Da98b954EedeAC495271d0F",
            "0x6B3595068778DD592e39A122f4f5a5cF09C90fE2",
            "0x6f259637dcD74C767781E37Bc6133cd6A68aa161",
            "0x75231F58b43240C9718Dd58B4967c5114342a86c",
            "0x7D1AfA7B718fb893dB30A3aBc0Cfc608AaCfeBB0",
            "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",
            "0x8290333ceF9e6D528dD5618Fb97a76f268f3EDD4",
            "0x84cA8bc7997272c7CfB4D0Cd3D55cd942B3c9419",
            "0x8Ab7404063Ec4DBcfd4598215992DC3F8EC853d7",
            "0x8CE9137d39326AD0cD6491fb5CC0CbA0e089b6A9",
            "0x967da4048cD07aB37855c090aAF366e4ce1b9F48",
            "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2",
            "0xA0b73E1Ff0B80914AB6fe0444E65848C4C34450b",
            "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
            "0xa3BeD4E1c75D00fa6f4E5E6922DB7261B5E9AcD2",
            "0xA9B1Eb5908CfC3cdf91F9B8B3a74108598009096",
            "0xADE00C28244d5CE17D72E40330B1c318cD12B7c3",
            "0xBA11D00c5f74255f56a5E366F4F77f5A186d7f55",
            "0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB",
            "0xBBbbCA6A901c926F240b89EacB641d8Aec7AEafD",
            "0xbE9375C6a420D2eEB258962efB95551A5b722803",
            "0xc00e94Cb662C3520282E6f5717214004A7f26888",
            "0xC011a73ee8576Fb46F5E1c5751cA3B9Fe0af2a6F",
            "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
            "0xc12eCeE46ed65D970EE5C899FCC7AE133AfF9b03",
            "0xC581b735A1688071A1746c968e0798D642EDE491",
            "0xcB3df3108635932D912632ef7132d03EcFC39080",
            "0xd26114cd6EE289AccF82350c8d8487fedB8A0C07",
            "0xD46bA6D942050d489DBd938a2C909A5d5039A161",
            "0xD533a949740bb3306d119CC777fa900bA034cd52",
            "0xD71eCFF9342A5Ced620049e616c5035F1dB98620",
            "0xdAC17F958D2ee523a2206206994597C13D831ec7",
            "0xdB25f211AB05b1c97D595516F45794528a807ad8",
            "0xdd974D5C2e2928deA5F71b9825b8b646686BD200",
            "0xe1aFe1Fd76Fd88f78cBf599ea1846231B8bA3B6B",
            "0xe28b3B32B6c345A34Ff64674606124Dd5Aceca30",
            "0xE41d2489571d322189246DaFA5ebDe1F4699F498",
            "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "0xf8C3527CC04340b208C854E985240c02F7B7793f",
            "0xfF20817765cB7f73d4bde2e66e067E58D11095C2",
        ],
        Network.Fantom: [
            "0x21be370D5312f44cB42ce377BC9b8a0cEF1A4C83",
            "0x321162Cd933E2Be498Cd2267a90534A804051b11",
            "0x2406dCe4dA5aB125A18295f4fB9FD36a0f7879A2",
            "0x74b23882a30290451A17c44f4F05243b6b58C76d",
            "0xBDC8fd437C489Ca3c6DA3B5a336D11532a532303",
            "0xd6070ae98b8069de6B494332d1A1a81B6179D960",
            "0x1E4F97b9f9F913c46F1632781732927B9019C68b",
            "0x6a07A792ab2965C72a5B8088d3a069A7aC3a993B",
            "0x657A1861c15A3deD9AF0B6799a195a249ebdCbc6",
            "0xb3654dc3D10Ea7645f8319668E8F54d2574FBdC8",
            "0x399fe752D39338d28C36F3370fbebd8292fb9E6e",
            "0x56ee926bD8c72B2d5fa1aF4d9E4Cbb515a1E3Adc",
            "0x468003B688943977e6130F4F68F23aad939a1040",
            "0xae75A438b2E0cB8Bb01Ec1E1e376De11D44477CC",
            "0x81740D647493a61329E1c574A11ee7577659fb14",
            "0xe105621721D1293c27be7718e041a4Ce0EbB227E",
            "0x29b0Da86e484E1C0029B56e817912d778aC0EC69",
        ],
    }.get(chain.id, [])
)
FEEDS = list(feeds)


@pytest.mark.parametrize("token", FEEDS)
@pytest.mark.asyncio_cooperative
async def test_chainlink_get_feed(token):
    """
    Test ``chainlink.get_feed`` with both lowercase and checksum addresses.

    This test verifies that ``chainlink.get_feed`` returns a valid Feed object by
    asserting that the returned object is not equal to ZERO_ADDRESS.

    See Also:
        :func:`y.prices.chainlink.chainlink.get_feed`
    """
    assert await chainlink.get_feed(token, sync=False) != ZERO_ADDRESS, "no feed available"


@pytest.mark.parametrize("token", FEEDS)
@pytest.mark.asyncio_cooperative
async def test_chainlink_latest(token):
    """
    Test the latest price availability for a token using Chainlink.

    This test checks whether ``chainlink.get_price(token)`` returns a valid price.
    If no price is returned, it retrieves the corresponding Feed object from
    ``chainlink.get_feed(token)`` and checks its latest timestamp. If the timestamp
    indicates that the feed is stale (i.e. more than 24 hours old relative to the current time),
    the test is skipped. Otherwise, it asserts that the associated feed's contract's
    aggregator property is the zero address, thereby indicating that no current price is provided.

    See Also:
        :func:`y.prices.chainlink.chainlink.get_feed`
    """
    if not await chainlink.get_price(token):
        feed = await chainlink.get_feed(token)
        latest_timestamp = await feed.latest_timestamp()
        if latest_timestamp and latest_timestamp + 24 * 60 * 60 < time.time():
            pytest.skip("feed is stale")
        try:
            assert await feed.contract.aggregator == ZERO_ADDRESS, "no current price available"
        except AttributeError as e:
            raise AttributeError(*e.args, feed) from e


@mainnet_only
@pytest.mark.parametrize("token", FEEDS)
@pytest.mark.asyncio_cooperative
async def test_chainlink_before_registry(token):
    """
    Test Chainlink price retrieval for a token at a block prior to the availability
    of the Chainlink registry.

    This test verifies that a price retrieval attempt using ``chainlink.get_price``
    at a historical block (set to 12800000) either returns a valid price or, if not,
    examines the Feed object. If the feed was deployed after the test block, the test
    is skipped. Otherwise, if no price is returned and the feed is not stale, it asserts
    that the feed's contract aggregator property equals the zero address, indicating
    that no price is available before the registry.

    See Also:
        :func:`y.prices.chainlink.chainlink.get_price`
    """
    test_block = 12800000
    assert chainlink.asynchronous is True
    feed = await chainlink.get_feed(token, sync=False)
    if await contract_creation_block_async(feed.address) > test_block:
        pytest.skip("not applicable to feeds deployed after test block")
    price = chainlink.get_price(token, block=test_block)
    price = await price
    if not price:
        feed = await chainlink.get_feed(token)
        latest_timestamp = await feed.latest_timestamp()
        if latest_timestamp and latest_timestamp + 24 * 60 * 60 < time.time():
            pytest.skip("feed is stale")
        try:
            assert (
                await feed.contract.aggregator == ZERO_ADDRESS
            ), f"{feed} no price available before registry"
        except AttributeError as e:
            raise AttributeError(*e.args, feed) from e


@pytest.mark.asyncio_cooperative
async def test_chainlink_nonexistent():
    """
    Test Chainlink Feed and price retrieval for a nonexistent token.

    This test verifies that attempting to retrieve a Feed or price using a nonexistent
    token (the zero address) correctly returns None.

    See Also:
        :func:`y.prices.chainlink.chainlink.get_feed`
    """
    assert await chainlink.get_feed(ZERO_ADDRESS, sync=False) is None
    assert await chainlink.get_price(ZERO_ADDRESS, sync=False) is None


@mainnet_only
@pytest.mark.asyncio_cooperative
async def test_chainlink_before_feed():
    """
    Test Chainlink price retrieval one block before a feed is deployed.

    This test checks that querying the price for a token (e.g., YFI) one block prior to
    its associated Chainlink Feed deployment returns None.

    See Also:
        :func:`y.prices.chainlink.chainlink.get_price`
    """
    # try to fetch yfi price one block before feed is deployed
    assert (
        await chainlink.get_price(
            "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e", 12742718, sync=False
        )
        is None
    )



UNIV3_QUOTER_ABI = [
    {
        "inputs": [
            {"internalType": "address", "name": "_factory", "type": "address"},
            {"internalType": "address", "name": "_WETH9", "type": "address"},
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "inputs": [],
        "name": "WETH9",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "factory",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes", "name": "path", "type": "bytes"},
            {"internalType": "uint256", "name": "amountIn", "type": "uint256"},
        ],
        "name": "quoteExactInput",
        "outputs": [{"internalType": "uint256", "name": "amountOut", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "address", "name": "tokenOut", "type": "address"},
            {"internalType": "uint24", "name": "fee", "type": "uint24"},
            {"internalType": "uint256", "name": "amountIn", "type": "uint256"},
            {"internalType": "uint160", "name": "sqrtPriceLimitX96", "type": "uint160"},
        ],
        "name": "quoteExactInputSingle",
        "outputs": [{"internalType": "uint256", "name": "amountOut", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes", "name": "path", "type": "bytes"},
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"},
        ],
        "name": "quoteExactOutput",
        "outputs": [{"internalType": "uint256", "name": "amountIn", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "tokenIn", "type": "address"},
            {"internalType": "address", "name": "tokenOut", "type": "address"},
            {"internalType": "uint24", "name": "fee", "type": "uint24"},
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"},
            {"internalType": "uint160", "name": "sqrtPriceLimitX96", "type": "uint160"},
        ],
        "name": "quoteExactOutputSingle",
        "outputs": [{"internalType": "uint256", "name": "amountIn", "type": "uint256"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "int256", "name": "amount0Delta", "type": "int256"},
            {"internalType": "int256", "name": "amount1Delta", "type": "int256"},
            {"internalType": "bytes", "name": "path", "type": "bytes"},
        ],
        "name": "uniswapV3SwapCallback",
        "outputs": [],
        "stateMutability": "view",
        "type": "function",
    },
]



import pytest
from brownie import chain
from multicall.utils import await_awaitable

from tests.fixtures import mainnet_only
from y.networks import Network
from y.prices.lending.aave import AaveRegistry

aDAI = "0x028171bCA77440897B824Ca71D1c56caC55b68A3"
DAI = "0x6B175474E89094C44Da98b954EedeAC495271d0F"

ATOKENS = {
    Network.Mainnet: [
        "0xfC1E690f61EFd961294b3e1Ce3313fBD8aa4f85d",
        "0x4DA9b813057D04BAef4e5800E36083717b4a0341",
        "0x9bA00D6856a4eDF4665BcA2C2309936572473B7E",
        "0x71fc860F7D3A592A4a98740e39dB31d25db65ae8",
        "0x625aE63000f46200499120B906716420bd059240",
        "0x7D2D3688Df45Ce7C552E19c27e007673da9204B8",
        "0xE1BA0FB44CCb0D11b80F92f4f8Ed94CA3fF51D00",
        "0x3a3A65aAb0dd2A17E3F1947bA16138cd37d08c04",
        "0xA64BD6C70Cb9051F6A9ba1F163Fdc07E0DfB5F84",
        "0x9D91BE44C06d373a8a226E1f3b146956083803eB",
        "0x71010A9D003445aC60C4e6A7017c1E89A477B438",
        "0x7deB5e830be29F91E298ba5FF1356BB7f8146998",
        "0x6FCE4A401B6B80ACe52baAefE4421Bd188e76F6f",
        "0x6Fb0855c404E09c47C3fBCA25f08d4E41f9F062f",
        "0x328C4c80BC7aCa0834Db37e6600A6c49E12Da4DE",
        "0xFC4B8ED459e00e5400be803A9BB3954234FD50e3",
        "0x6Ee0f7BB50a54AB5253dA0667B0Dc2ee526C30a8",
        "0x712DB54daA836B53Ef1EcBb9c6ba3b9Efb073F40",
        "0x69948cC03f478B95283F7dbf1CE764d0fc7EC54C",
        "0x12e51E77DAAA58aA0E9247db7510Ea4B46F9bEAd",
        "0xba3D9687Cf50fE253cd2e1cFeEdE1d6787344Ed5",
        "0xB124541127A0A657f056D9Dd06188c4F1b0e5aab",
        "0x3Ed3B47Dd13EC9a98b44e6204A523E766B225811",
        "0x9ff58f4fFB29fA2266Ab25e75e2A8b3503311656",
        "0x030bA81f1c18d280636F32af80b9AAd02Cf0854e",
        "0x5165d24277cD063F5ac44Efd447B27025e888f37",
        "0xDf7FF54aAcAcbFf42dfe29DD6144A69b629f8C9e",
        "0xB9D7CB55f463405CDfBe4E90a6D2Df01C2B92BF1",
        "0xFFC97d72E13E01096502Cb8Eb52dEe56f74DAD7B",
        "0x05Ec93c0365baAeAbF7AefFb0972ea7ECdD39CF1",
        "0xA361718326c15715591c299427c62086F69923D9",
        "0x028171bCA77440897B824Ca71D1c56caC55b68A3",
        "0xaC6Df26a590F08dcC95D5a4705ae8abbc88509Ef",
        "0x39C6b3e42d6A679d7D776778Fe880BC9487C2EDA",
        "0xa06bC25B5805d5F8d82847D191Cb4Af5A3e873E0",
        "0xa685a61171bb30d4072B338c80Cb7b2c865c873E",
        "0xc713e5E149D5D0715DcD1c156a020976e7E56B88",
        "0xCC12AbE4ff81c9378D670De1b57F8e0Dd228D77a",
        "0x35f6B052C598d933D69A4EEC4D04c73A191fE6c2",
        "0x6C5024Cd4F8A59110119C56f8933403A539555EB",
        "0x101cc05f4A51C0319f570d5E146a8C625198e636",
        "0xBcca60bB61934080951369a648Fb03DF4F96263C",
        "0x8dAE6Cb04688C62d939ed9B68d32Bc62e49970b1",
        "0xD37EE7e4f452C6638c96536e68090De8cBcdb583",
        "0x272F97b7a56a387aE942350bBC7Df5700f8a4576",
        "0xF256CC7847E919FAc9B808cC216cAc87CCF2f47a",
        "0x514cd6756CCBe28772d4Cb81bC3156BA9d1744aa",
        "0xc9BC48c72154ef3e5425641a3c747242112a46AF",
        "0x1E6bb68Acec8fefBD87D192bE09bb274170a0548",
        "0x2e8F4bdbE3d47d7d7DE490437AeA9915D930F1A3",
        "0x6F634c6135D2EBD550000ac92F494F9CB8183dAe",
        "0xd4937682df3C8aEF4FE912A96A74121C0829E664",
        "0x683923dB55Fead99A79Fa01A27EeC3cB19679cC3",
        "0x1982b2F5814301d4e9a8b0201555376e62F82428",
        "0x9a14e23A58edf4EFDcB360f68cd1b95ce2081a2F",
        "0xc2e2152647F4C26028482Efaf64b2Aa28779EFC4",
        "0xf9Fb4AD91812b704Ba883B11d2B576E890a6730A",
        "0x79bE75FFC64DD58e66787E4Eae470c8a1FD08ba4",
        "0xd24946147829DEaA935bE2aD85A3291dbf109c80",
        "0x17a79792Fe6fE5C95dFE95Fe3fCEE3CAf4fE4Cb7",
        "0x13B2f6928D7204328b0E8E4BCd0379aA06EA21FA",
        "0x9303EabC860a743aABcc3A1629014CaBcc3F8D36",
        "0xc58F53A8adff2fB4eb16ED56635772075E2EE123",
        "0xe59d2FF6995a926A574390824a657eEd36801E55",
        "0xA1B0edF4460CC4d8bFAA18Ed871bFF15E5b57Eb4",
        "0xE340B25fE32B1011616bb8EC495A4d503e322177",
        "0x0ea20e7fFB006d4Cfe84df2F72d8c7bD89247DB0",
        "0xb8db81B84d30E2387de0FF330420A4AAA6688134",
        "0x370adc71f67f581158Dc56f539dF5F399128Ddf9",
        "0xA9e201A4e269d6cd5E9F0FcbcB78520cf815878B",
        "0x38E491A71291CD43E8DE63b7253E482622184894",
        "0x3D26dcd840fCC8e4B2193AcE8A092e4a65832F9f",
        "0x391E86e2C002C70dEe155eAceB88F7A3c38f5976",
        "0x2365a4890eD8965E564B7E2D27C38Ba67Fec4C6F",
        "0x5394794Be8b6eD5572FCd6b27103F46b5F390E8f",
        "0x358bD0d980E031E23ebA9AA793926857703783BD",
        "0xd109b2A304587569c84308c55465cd9fF0317bFB",
        "0xd145c6ae8931ed5Bca9b5f5B7dA5991F5aB63B5c",
        "0xCa5DFDABBfFD58cfD49A9f78Ca52eC8e0591a3C5",
        "0x2394eAc76A0B7992FB8aa0F0C126317Af76Ae707",
        "0x0EDf29d931b68D673bdCDD35B20e26eEC0532F25",
        "0x762E5982EB0D944A9800D7caBAF640464E892C91",
        "0x50D695cf7a86cFe415642cBCb25B025a9B1597e7",
        "0x0A6172fb19E85DBae18898eA2f7a596B5681f1eE",
    ]
}.get(chain.id, [])

supported_chains = pytest.mark.skipif(
    not (AaveRegistry(asynchronous=False).pools or ATOKENS),
    reason="Not applicable on chains without known Aaves or forks",
)


@mainnet_only
def test_adai():
    """
    Simple test for a known underlying.

    This test verifies that the underlying asset of a known aToken (aDAI) is correctly identified as DAI.

    See Also:
        - :func:`AaveRegistry.underlying`
    """
    assert AaveRegistry(asynchronous=False).underlying(aDAI) == DAI
    assert await_awaitable(AaveRegistry(asynchronous=True).underlying(aDAI)) == DAI


@supported_chains
def test_AaveRegistry():
    """
    Test the AaveRegistry for pool and aToken availability.

    This test checks that the AaveRegistry can fetch pools and that the `ATOKENS` constant is set up correctly.

    See Also:
        - :class:`AaveRegistry`
    """
    aave = AaveRegistry()
    assert aave.pools, f"Cannot fetch Aave pools on {Network.printable()}"
    assert ATOKENS, f"Please set up `ATOKENS` constant to test Aave on {Network.printable()}"
    for pool in aave.pools:
        assert pool.atokens, f"Cannot find atokens for Aave pool {pool}"


@pytest.mark.parametrize("token", ATOKENS)
@pytest.mark.asyncio_cooperative
async def test_atokens_async(token):
    """
    Test various functions related to aTokens asynchronously.

    This test verifies the following functions:
    - :meth:`AaveRegistry.is_atoken`
    - :meth:`AaveRegistry.pool_for_atoken`
    - :meth:`AaveRegistry.underlying`
    - :meth:`AaveRegistry.get_price`

    Args:
        token: The aToken address to test.

    See Also:
        - :class:`AaveRegistry`
    """
    aave = AaveRegistry(asynchronous=True)
    with pytest.raises(RuntimeError):
        assert token in aave, f"Cannot find atoken {token} in AaveRegistry."

    assert await aave.is_atoken(token), f"Cannot validate atoken {token} as an atoken."
    assert await aave.pool_for_atoken(token), f"Cannot find pool for atoken {token}"
    underlying = await aave.underlying(token)
    assert underlying, f"Cannot find underlying for atoken {token}"
    assert await aave.get_price(token), f"Cannot find price for atoken {token}"


@pytest.mark.asyncio_cooperative
async def test_wrapped_atoken_v2():
    """
    Test for wrapped aToken v2 pricing.

    This test checks if a wrapped aToken v2 is correctly identified and priced.

    See Also:
        - :meth:`AaveRegistry.is_wrapped_atoken_v2`
        - :meth:`AaveRegistry.get_price_wrapped_v2`
    """
    wrapped_ausdt = "0xf8Fd466F12e236f4c96F7Cce6c79EAdB819abF58"
    aave: AaveRegistry = AaveRegistry(asynchronous=True)
    assert await aave.is_wrapped_atoken_v2(wrapped_ausdt)
    assert await aave.get_price_wrapped_v2(wrapped_ausdt, block=17_000_000) == 1.112026


@pytest.mark.asyncio_cooperative
async def test_wrapped_atoken_v3():
    """
    Test for wrapped aToken v3 pricing.

    This test checks if a wrapped aToken v3 is correctly identified and priced.

    See Also:
        - :meth:`AaveRegistry.is_wrapped_atoken_v3`
        - :meth:`AaveRegistry.get_price_wrapped_v3`
    """
    wrapped_ausdc = "0x57d20c946A7A3812a7225B881CdcD8431D23431C"
    aave: AaveRegistry = AaveRegistry(asynchronous=True)
    assert await aave.is_wrapped_atoken_v3(wrapped_ausdc)
    assert await aave.get_price_wrapped_v3(wrapped_ausdc, block=17_000_000) == 1.000168



from decimal import Decimal
from typing import Optional

from a_sync import a_sync
from dank_mids.exceptions import Revert
from multicall.call import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, WeiBalance
from y.contracts import Contract, has_methods
from y.datatypes import Address, Block
from y.utils.cache import optional_async_diskcache


METHODS = "main()(address)", "issuanceAvailable()(uint)", "redemptionAvailable()(uint)"


@a_sync(default="sync")
@optional_async_diskcache
async def is_rtoken(token_address: Address) -> bool:
    """
    Determine whether a given token is a Reserve Protocol R-token.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a Reserve Protocol R-token, False otherwise.

    Example:
        >>> address = "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"  # USDC (not an R-token)
        >>> result = is_rtoken(address)
        >>> print(result)
        False

    See Also:
        :func:`~y.prices.tokenized_fund.reserve.get_price`
    """
    return await has_methods(token_address, METHODS, sync=False)


@a_sync(default="sync")
async def get_price(
    token_address: Address,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> Decimal:
    """
    Retrieve the USD price of a Reserve Protocol R-token.

    This function obtains the basket handler associated with the R-token by calling its
    "main" and "basketHandler" methods. It then calls the basket handler's price method to retrieve a
    low and a high price value. The function computes the integer average of these two price values and scales
    the result by 1e18 to convert it to a USD price. If the call to the basket handler's price method reverts,
    the function returns None instead of raising an exception.

    Args:
        token_address: The address of the R-token.
        block: The block number to query. Defaults to the latest block if not specified.
        skip_cache: If True, skip using the cache. Defaults to :attr:`~y.ENVIRONMENT_VARIABLES.SKIP_CACHE`.

    Returns:
        The USD price of the R-token as computed from the basket handler, or None if the price cannot
        be determined due to a remote call failure (i.e. a :class:`~dank_mids.exceptions.Revert`).

    Raises:
        TypeError: If the token is not a valid R-token.

    Example:
        >>> address = "0xaCeeD87BD5754c3d714F3Bd43a9B7B0C9250ab0D"  # RSV token
        >>> price = await get_price(address, sync=False)
        >>> if price is None:
        ...     print("Price calculation failed due to a Revert exception")
        ... else:
        ...     print(f"Calculated price: {price}")
        >>> # In a synchronous context:
        >>> price = get_price(address)
        >>> print(price)

    See Also:
        :func:`~y.prices.tokenized_fund.reserve.is_rtoken`
    """
    main = await Call(token_address, "main()(address)", block_id=block)
    if main is None:
        raise TypeError(main, token_address, await is_rtoken(token_address))
    basket_handler = await Contract.coroutine(
        await Call(main, "basketHandler()(address)", block_id=block)
    )
    try:
        low, high = await basket_handler.price.coroutine(block_identifier=block)
    except Revert:
        return None
    else:
        return Decimal(low + high) // 2 / 10**18

    tokens, *_ = await basket_handler.getPrimeBasket.coroutine(block_identifier=block)
    tokens = [ERC20(token, asynchronous=True) for token in tokens]
    balances = [
        WeiBalance(balance, token, block=block, skip_cache=skip_cache, asynchronous=True)
        for token, balance in zip(
            tokens, await basket_handler.quantity.map(tokens, block_identifier=block)
        )
    ]
    print(balances)
    values = WeiBalance.value_usd.map(balances).values()
    print(values)
    value = sum(values)
    print(value)
    supply = await ERC20(token_address, asynchronous=True).total_supply_readable(block)
    print(f"ts: {supply}")
    raise Exception(value / supply)



from abc import ABCMeta, abstractmethod
from asyncio import Task, create_task, get_event_loop, sleep
from copy import copy
from itertools import dropwhile, groupby
from logging import DEBUG, getLogger
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Container,
    Dict,
    Generic,
    List,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

import a_sync
import dank_mids
import eth_retry
from a_sync import (
    ASyncIterable,
    ASyncIterator,
    AsyncThreadPoolExecutor,
    CounterLock,
    PruningThreadPoolExecutor,
)
from async_property import async_property
from brownie import ZERO_ADDRESS
from dank_mids import BlockSemaphore
from evmspec.data import Address, HexBytes32
from hexbytes import HexBytes
from pony.orm import OptimisticCheckError, TransactionIntegrityError, db_session
from web3.datastructures import AttributeDict
from web3.middleware.filter import block_ranges

from y import convert
from y import ENVIRONMENT_VARIABLES as ENVS
from y._db.decorators import retry_locked
from y._db.exceptions import CacheNotPopulatedError
from y._decorators import stuck_coro_debugger
from y.exceptions import reraise_excs_with_extra_context
from y.utils.middleware import BATCH_SIZE

if TYPE_CHECKING:
    from y import Block

T = TypeVar("T")
S = TypeVar("S")
M = TypeVar("M")

Checkpoints = Dict["Block", int]

logger = getLogger(__name__)
default_filter_threads = PruningThreadPoolExecutor(4)
"""
The thread pool executor used for all :class:`Filter` objects without one provided, with a maximum of 4 threads.
"""


def enc_hook(obj: Any) -> bytes:
    """
    Encode hook for JSON serialization of special types.

    Args:
        obj: The object to encode.

    Raises:
        NotImplementedError: If the object type is not supported for encoding.

    Note:
        Currently supports encoding of :class:`int`, :class:`Address`, :class:`HexBytes32`, :class:`HexBytes`, and :class:`AttributeDict` objects.

    Examples:
        >>> from web3.datastructures import AttributeDict
        >>> enc_hook(AttributeDict({'key': 'value'}))
        {'key': 'value'}

        >>> from hexbytes import HexBytes
        >>> enc_hook(HexBytes('0x1234'))
        '1234'

    See Also:
        - :func:`dec_hook`
    """
    typ = type(obj)
    # sometimes we get a recursion error from the instance checks, this helps us debug that case.
    with reraise_excs_with_extra_context(obj, typ):
        # we use issubclass instead of isinstance here to prevent a recursion error
        if issubclass(typ, int):
            return int(obj)
        elif issubclass(typ, Address):
            return obj[2:]
        elif isinstance(obj, HexBytes32):
            # we trim all leading zeroes since we know how many we need to put back later
            return hex(int(obj.hex(), 16))[2:]
        elif isinstance(obj, HexBytes):
            return bytes(obj).hex()
        elif isinstance(obj, AttributeDict):
            return dict(obj)
        raise TypeError


def dec_hook(typ: Type[T], obj: bytes) -> T:
    """
    Decode hook for JSON deserialization of special types.

    Args:
        typ: The type to decode into.
        obj: The object to decode.

    Raises:
        ValueError: If the type is not supported for decoding.

    Note:
        Currently only supports decoding of :class:`HexBytes` objects.

    Example:
        >>> from hexbytes import HexBytes
        >>> dec_hook(HexBytes, b'1234')
        HexBytes('0x1234')

    See Also:
        - :func:`enc_hook`
    """
    if typ is HexBytes:
        return typ(obj)
    raise ValueError(f"{typ} is not a valid type for decoding")


class DiskCache(Generic[S, M], metaclass=ABCMeta):
    @abstractmethod
    def _set_metadata(self, from_block: "Block", done_thru: "Block") -> None:
        """
        Update cache metadata to indicate that the cache is populated from block `from_block` to block `done_thru`.

        Args:
            from_block: The starting block number.
            done_thru: The ending block number indicating that the cache is populated up to this block.

        Example:
            >>> disk_cache._set_metadata(100, 200)

        See Also:
            - :meth:`set_metadata`
        """

    @abstractmethod
    def _is_cached_thru(self, from_block: "Block") -> "Block":
        """
        Returns max cached block for this cache or 0 if not cached.

        Args:
            from_block: The starting block number.

        Returns:
            The maximum cached block number.
        """

    @abstractmethod
    def _select(self, from_block: "Block", to_block: "Block") -> List[S]:
        """
        Selects all cached objects from block `from_block` to block `to_block`.

        Args:
            from_block: The starting block number.
            to_block: The ending block number.

        Returns:
            A list of cached objects.
        """

    @retry_locked
    def set_metadata(self, from_block: "Block", done_thru: "Block") -> None:
        """
        Update cache metadata to indicate that the cache is populated from block `from_block` to block `done_thru`.

        Args:
            from_block: The starting block number.
            done_thru: The ending block number up to which the cache is populated.

        Example:
            >>> cache.set_metadata(100, 200)

        See Also:
            - :meth:`_set_metadata`
            - :class:`CacheNotPopulatedError`
        """
        try:
            with db_session:
                self._set_metadata(from_block, done_thru)
        except TransactionIntegrityError as e:
            logger.debug("%s got exc %s when setting cache metadata", self, e)
            self.set_metadata(from_block, done_thru)
        except OptimisticCheckError as e:
            # Don't need to update in this case
            logger.debug("%s got exc %s when setting cache metadata", self, e)

    @db_session
    @retry_locked
    def select(self, from_block: "Block", to_block: "Block") -> List[S]:
        """
        Selects all cached objects from block `from_block` to block `to_block`.

        Args:
            from_block: The starting block number.
            to_block: The ending block number.

        Returns:
            A list of cached objects.
        """
        return self._select(from_block, to_block)

    @db_session
    @retry_locked
    def is_cached_thru(self, from_block: "Block") -> "Block":
        """
        Returns max cached block for this cache or 0 if not cached.

        Args:
            from_block: The starting block number.

        Returns:
            The maximum cached block number.
        """
        return self._is_cached_thru(from_block)

    @db_session
    @retry_locked
    def check_and_select(self, from_block: "Block", to_block: "Block") -> List[S]:
        """
        Selects all cached objects within a specified block range.

        Args:
            from_block: The starting block number.
            to_block: The ending block number.

        Returns:
            A list of cached objects.

        Raises:
            CacheNotPopulatedError: If the cache is not fully populated.

        Example:
            >>> try:
            ...     data = cache.check_and_select(100, 200)
            ... except CacheNotPopulatedError:
            ...     print("Cache incomplete")

        See Also:
            - :meth:`set_metadata`
            - :meth:`select`
        """
        if self.is_cached_thru(from_block) >= to_block:
            return self.select(from_block, to_block)
        raise CacheNotPopulatedError(self, from_block, to_block)

    __slots__ = []


C = TypeVar("C", bound=DiskCache)


class _DiskCachedMixin(ASyncIterable[T], Generic[T, C], metaclass=ABCMeta):
    """
    Mixin that provides asynchronous features for data caches stored on disk.
    """

    _checkpoints: Checkpoints
    __slots__ = "is_reusable", "_cache", "_executor", "_objects", "_pruned"

    def __init__(
        self,
        executor: Optional[AsyncThreadPoolExecutor] = None,
        is_reusable: bool = True,
    ):
        """
        Initializes the mixin.

        Args:
            executor: Optional executor to use for I/O operations.
            is_reusable: Whether data should be kept in memory for reuse.
        """
        self.is_reusable = is_reusable
        self._cache = None
        self._executor = executor
        self._objects: List[T] = []
        self._pruned = 0

    @property
    @abstractmethod
    def cache(self) -> C:
        """
        Returns the associated cache object, which must be defined in subclasses.
        """

    @property
    def executor(self) -> AsyncThreadPoolExecutor:
        """
        Returns the executor used for disk operations, creating one if necessary.
        """
        executor = self._executor
        if executor is None:
            executor = self._executor = AsyncThreadPoolExecutor(1)
        return executor

    def __del__(self) -> None:
        """
        Shutdown the executor on deletion.
        """
        executor = self._executor
        if executor is not None:
            executor.shutdown()

    @property
    @abstractmethod
    def insert_to_db(self) -> Callable[[T], None]: ...

    def bulk_insert(self) -> Callable[[List[T]], Awaitable[None]]:
        """
        Function to bulk insert a list of objects into the database.

        This property must be overridden in subclasses to provide the desired bulk insertion logic.
        The implementation should return a callable that accepts a list of objects (List[T])
        and returns an awaitable.

        Example:
            >>> async def my_bulk_insert(objs):
            ...     # perform custom bulk insert operations here
            ...     pass
            >>> class MyFilter(_DiskCachedMixin):
            ...     @property
            ...     def bulk_insert(self) -> Callable[[List[T]], Awaitable[None]]:
            ...         return my_bulk_insert

        See Also:
            - :meth:`_load_cache`
        """

    async def _extend(self, objs: Container[T]) -> None:
        """
        Override this to pre-process objects before storing.

        Args:
            objs ("Container[T]"): The objects to extend the list with.

        Example:
            >>> await instance._extend([obj1, obj2])

        See Also:
            - :meth:`_load_cache`
        """
        if objs:
            self._objects.extend(objs)
            if self.is_reusable:
                block = self._get_block_for_obj(self._objects[-1])
                self._checkpoints[block] = len(self._objects)

    async def _load_cache(self, from_block: "Block") -> "Block":
        """
        Loads cached logs from disk.

        Args:
            from_block: The starting block number.

        Returns:
            The maximum block number loaded from cache, or None if no cached data is available.

        Example:
            >>> cached_thru = await instance._load_cache(100)
            >>> if cached_thru is None:
            ...     print("No cached data available")
            ... else:
            ...     print(f"Cache loaded through block {cached_thru}")

        See Also:
            - :meth:`_extend`
        """
        logger.debug("checking to see if %s is cached in local db", self)
        if cached_thru := await _metadata_read_executor.run(self.cache.is_cached_thru, from_block):
            logger.info("%s is cached thru block %s, loading from db", self, cached_thru)
            await self._extend(await self.executor.run(self.cache.select, from_block, cached_thru))
            if self.is_reusable:
                objs_per_chunk = 50
                num_checkpoints = len(self._objects) // objs_per_chunk
                checkpoint_indexes = (i * objs_per_chunk for i in range(1, num_checkpoints))
                get_block_for_obj = self._get_block_for_obj
                for index in checkpoint_indexes:
                    obj = self._objects[index]
                    if index < len(self._objects):
                        next_obj = self._objects[index + 1]
                        while get_block_for_obj(obj) == get_block_for_obj(next_obj):
                            obj = next_obj
                            index += 1
                            try:
                                next_obj = self._objects[index + 1]
                            except IndexError:
                                break

                    self._checkpoints[get_block_for_obj(obj)] = index

            logger.info(
                "%s loaded %s objects thru block %s from disk",
                self,
                len(self._objects),
                cached_thru,
            )
            return cached_thru
        return None


def make_executor(small: int, big: int, name: Optional[str] = None) -> PruningThreadPoolExecutor:
    """
    Creates a thread pool executor that prunes completed tasks.

    Args:
        small: Size of the pool if using a non-postgres DB.
        big: Size of the pool if using a postgres DB.
        name: Optional name for the executor.

    Returns:
        A PruningThreadPoolExecutor instance based on the environment configuration.
    """
    return PruningThreadPoolExecutor(big if ENVS.DB_PROVIDER == "postgres" else small, name)


_E = TypeVar("_E", bound=AsyncThreadPoolExecutor)
_MAX_LONG_LONG = 9223372036854775807

_metadata_read_executor = make_executor(2, 3, "ypricemagic Filter read metadata")
_metadata_write_executor = make_executor(1, 3, "ypricemagic Filter write metadata")


class Filter(_DiskCachedMixin[T, C]):
    # defaults are stored as class vars to keep instance dicts smaller
    _chunk_size = BATCH_SIZE
    _chunks_per_batch = None
    _exc = None
    _tb = None
    _db_task = None
    _sleep_fut = None
    _sleep_time = 60
    _task = None
    _depth = 0
    _semaphore = None
    _verbose = False
    __slots__ = (
        "from_block",
        "to_block",
        "_checkpoints",
        "_interval",
        "_lock",
    )

    def __init__(
        self,
        from_block: "Block",
        *,
        chunk_size: int = BATCH_SIZE,
        chunks_per_batch: Optional[int] = None,
        sleep_time: int = 60,
        semaphore: Optional[BlockSemaphore] = None,
        executor: Optional[AsyncThreadPoolExecutor] = None,
        is_reusable: bool = True,
        verbose: bool = False,
    ):
        """
        Initializes the Filter, specifying blocks, concurrency, and caching parameters.

        Args:
            from_block: Earliest block from which data should be retrieved.
            chunk_size: Size of each chunk to retrieve in a single RPC call.
            chunks_per_batch: How many chunks to load per batch (optional).
            sleep_time: Time (in seconds) to sleep between reloads when following chain head.
            semaphore: Block-based semaphore to limit concurrency.
            executor: Executor for disk-related tasks.
            is_reusable: Keeps data in memory if True; data is pruned when False.
            verbose: Enable debug-like progress logging if True.
        """
        self.from_block = from_block
        if chunk_size != self._chunk_size:
            self._chunk_size = chunk_size
        if chunks_per_batch != self._chunks_per_batch:
            self._chunks_per_batch = chunks_per_batch
        self._lock = CounterLock(name=str(self))
        if semaphore != self._semaphore:
            self._semaphore = semaphore
        if sleep_time != self._sleep_time:
            self._sleep_time = sleep_time
        if verbose != self._verbose:
            self._verbose = verbose
        self._checkpoints = {}
        super().__init__(executor=executor, is_reusable=is_reusable)

    def __aiter__(self) -> AsyncIterator[T]:
        """
        Returns an async iterator over the stored objects, yielding new ones as they are fetched.
        """
        return self._objects_thru(block=None).__aiter__()

    def __del__(self) -> None:
        """
        Cancels any pending fetch task upon deletion.
        """
        if self._task and not self._task.done():
            self._task.cancel()

    @abstractmethod
    async def _fetch_range(self, from_block: "Block", to_block: "Block") -> List[T]:
        """
        Fetches data for a given range of blocks from an on-chain or remote provider.

        Args:
            from_block: Lower bound of the block range.
            to_block: Upper bound of the block range.

        Returns:
            List of objects for that block range.
        """

    @property
    def semaphore(self) -> BlockSemaphore:
        if self._semaphore is None:
            self._semaphore = BlockSemaphore(self._chunks_per_batch)
        return self._semaphore

    def _get_block_for_obj(self, obj: T) -> "Block":
        """
        Override this as needed for different object types.

        Args:
            obj: The object to get the block number for.

        Returns:
            The block number of the object.

        Example:
            >>> block = instance._get_block_for_obj(some_obj)
        """
        return obj.blockNumber

    @ASyncIterator.wrap
    async def _objects_thru(
        self, block: Optional["Block"], from_block: Optional["Block"] = None
    ) -> AsyncIterator[T]:
        """
        Generates objects up to a specified block, or indefinitely if none is given.

        Args:
            block: Maximum block number to generate objects. If None, yields continuously.
            from_block: Minimum block from which to start yielding objects. Only valid if reusable.

        Yields:
            Objects that fall within the requested block range, if any.
        """
        self._ensure_task()
        debug_logs = logger.isEnabledFor(DEBUG)
        yielded = self._pruned
        done_thru = 0
        get_block_for_obj = self._get_block_for_obj
        if self.is_reusable:
            if from_block:
                reached_from_block = False

                def obj_out_of_range(obj) -> bool:
                    if get_block_for_obj(obj) < from_block:
                        return True
                    nonlocal reached_from_block
                    reached_from_block = True
                    return False

                def skip_too_early(objects):
                    nonlocal yielded
                    if checkpoints := self._checkpoints:
                        start_checkpoint_index = _get_checkpoint_index(from_block, checkpoints)
                        if start_checkpoint_index is not None:
                            objects = objects[start_checkpoint_index:]
                            yielded += start_checkpoint_index
                    start_len = len(objects)
                    objects = tuple(dropwhile(obj_out_of_range, objects))
                    yielded += start_len - len(objs)
                    return objects

            if objs := self._objects:
                if block is None:
                    if from_block:
                        objs = skip_too_early(objs)
                    for obj in objs:
                        yield obj
                    yielded += len(objs)
                    done_thru = get_block_for_obj(obj)
                elif self._checkpoints:
                    checkpoint_index = _get_checkpoint_index(block, self._checkpoints)
                    if checkpoint_index is not None:
                        objs = objs[:checkpoint_index]
                        done_thru = get_block_for_obj(objs[-1])
                        if from_block:
                            objs = skip_too_early(objs)
                        for obj in objs:
                            yield obj
                        yielded += len(objs)

                elif from_block:
                    skip_too_early(objs)

        elif from_block:
            raise RuntimeError(
                f"You cannot pass a value for `from_block` unless the {type(self).__name__} is reusable"
            )

        while True:
            if block is None or done_thru < block:
                self._wakeup()
                await self._lock.wait_for(done_thru + 1)
            if self._exc is not None:
                # raise a copy of it so multiple waiters don't destroy the traceback
                raise self._exc.with_traceback(self._tb) from self._exc.__cause__
            if to_yield := self._objects[yielded - self._pruned :]:
                if from_block and not reached_from_block:
                    objs = skip_too_early(to_yield)
                    if block is None:
                        for obj in objs:
                            yield obj
                    else:
                        for obj in objs:
                            if get_block_for_obj(obj) > block:
                                return
                            yield obj
                    yielded += len(objs)

                elif block:
                    if self.is_reusable:
                        for obj in to_yield:
                            if get_block_for_obj(obj) > block:
                                return
                            yield obj
                        yielded += len(to_yield)
                    else:
                        for obj in to_yield:
                            if get_block_for_obj(obj) > block:
                                self._prune(yielded - self._pruned)
                                return
                            yield obj
                            yielded += 1

                else:
                    for obj in to_yield:
                        yield obj
                    yielded += len(to_yield)

                if not self.is_reusable:
                    self._prune(len(to_yield))

            elif block and done_thru >= block:
                return

            done_thru = self._lock.value
            if debug_logs:
                logger._log(
                    DEBUG,
                    "%s lock value %s to_block %s",
                    (self, done_thru, block),
                )
            if block is None:
                await sleep(self._sleep_time)

    @async_property
    async def _sleep(self) -> None:
        """
        Puts the Filter into a sleep state until `_wakeup` is called. No new requests will be made.
        """
        if self._sleep_fut is None or self._sleep_fut.done():
            self._sleep_fut = get_event_loop().create_future()
        await self._sleep_fut

    def _wakeup(self) -> None:
        """Wake up the Filter to query logs from blocks not yet loaded into memory."""
        if self._sleep_fut is not None:
            self._sleep_fut.set_result(None)
            del self._sleep_fut

    async def __fetch(self) -> NoReturn:
        """
        Main coroutine that continuously runs the internal fetch loop.
        """
        try:
            await self._fetch()
        except Exception as e:
            import traceback

            logger.exception(e)
            self._exc = e
            self._tb = e.__traceback__
            # no need to hold vars in memory
            self._lock.set(_MAX_LONG_LONG)
            raise

    async def _fetch(self) -> NoReturn:
        """
        Defines the main logic for populating the Filter with data. Subclasses can override if needed.

        Example:
            >>> await instance._fetch()

        See Also:
            - :meth:`_loop`
        """
        await self._loop(self.from_block)

    @stuck_coro_debugger
    async def _fetch_range_wrapped(
        self, i: int, range_start: "Block", range_end: "Block", debug_logs: bool
    ) -> List[T]:
        """
        Wraps the _fetch_range call with concurrency control.

        Args:
            i: Index of the chunk or range segment.
            range_start: Lower bound of this block range.
            range_end: Upper bound of this block range.
            debug_logs: Whether debug logging is enabled.

        Returns:
            A tuple containing the index, the ending block, and the fetched objects.
        """
        async with self.semaphore[range_end]:
            if debug_logs:
                logger._log(
                    DEBUG,
                    "fetching %s block %s to %s",
                    (self, range_start, range_end),
                )
            return i, range_end, await self._fetch_range(range_start, range_end)

    async def _loop(self, from_block: "Block") -> NoReturn:
        """
        Work loop that continually fetches new data, loads from cache if available, then sleeps.

        Args:
            from_block: Earliest block from which to begin loading data.
        """
        logger.debug("starting work loop for %s", self)
        if cached_thru := await self._load_cache(from_block):
            self._lock.set(cached_thru)
        while True:
            await self._load_new_objects(start_from_block=cached_thru or from_block)
            await self._sleep

    @eth_retry.auto_retry
    @stuck_coro_debugger
    async def _load_new_objects(
        self,
        to_block: Optional["Block"] = None,
        start_from_block: Optional["Block"] = None,
    ) -> None:
        """
        Asynchronously loads new objects from your RPC, up to an optionally-specified end block.

        Args:
            to_block: Specific block to stop at. If None, load up to the current chain head.
            start_from_block: Block to start from if no prior data is cached.
        """
        SLEEP_TIME = 1

        if debug_logs := logger.isEnabledFor(DEBUG):
            logger._log(DEBUG, "loading new objects for %s", (self,))

        start = v + 1 if (v := self._lock.value) else start_from_block or self.from_block
        if to_block:
            end = to_block
            if start > end:
                raise ValueError(f"start {start} is bigger than end {end}, can't do that")

            while end > (current_block := await dank_mids.eth.block_number):
                logger.warning(
                    "You're trying to query a block range that has not fully completed:\n"
                    "range end: %s  current block: %s  Waiting 1s and trying again...",
                    end,
                    current_block,
                )
                await sleep(5.0)

        elif debug_logs:
            while start > (end := await dank_mids.eth.block_number):
                logger._log(
                    DEBUG,
                    "%s start %s is greater than end %s, sleeping...",
                    (self, start, end),
                )
                await sleep(SLEEP_TIME)

        else:
            while start > (end := await dank_mids.eth.block_number):
                await sleep(SLEEP_TIME)

        try:
            await self._load_range(start, end)
        except ValueError as e:
            if (
                "One of the blocks specified in filter (fromBlock, toBlock or blockHash) cannot be found."
                in str(e)
            ):
                logger.warning("Your rpc might be out of sync, trying again...")
            else:
                raise

    @stuck_coro_debugger
    async def _load_range(self, from_block: "Block", to_block: "Block") -> None:
        """
        Loads a particular block range in chunks, respecting concurrency limits.

        Args:
            from_block: Lower bound of the block range.
            to_block: Upper bound of the block range.
        """
        if debug_logs := logger.isEnabledFor(DEBUG):
            logger._log(DEBUG, "loading block range %s to %s", (from_block, to_block))
        chunks_yielded = 0
        done = {}
        coros = [
            self._fetch_range_wrapped(i, start, end, debug_logs)
            for i, (start, end) in enumerate(block_ranges(from_block, to_block, self._chunk_size))
            if self._chunks_per_batch is None or i < self._chunks_per_batch
        ]
        async for i, end, objs in a_sync.as_completed(coros, aiter=True, tqdm=self._verbose):
            next_chunk_loaded = False
            done[i] = end, objs
            for i in range(chunks_yielded, len(coros)):
                if i not in done:
                    break
                end, objs = done.pop(i)
                self._insert_chunk(objs, from_block, end, debug_logs)
                await self._extend(objs)
                next_chunk_loaded = True
                chunks_yielded += 1
            if next_chunk_loaded:
                await self._set_lock(end)
                if debug_logs:
                    logger._log(DEBUG, "%s loaded thru block %s", (self, end))

    @stuck_coro_debugger
    async def _set_lock(self, block: "Block") -> None:
        """
        Override this if you want to, for things like awaiting for tasks to complete as I do in the curve module.

        Args:
            block: The block number to set the lock to.

        Example:
            >>> await instance._set_lock(150)

        See Also:
            - :meth:`_load_new_objects`
        """
        self._lock.set(block)

    def _insert_chunk(
        self, objs: List[T], from_block: "Block", done_thru: "Block", debug_logs: bool
    ) -> None:
        """
        Queues the insertion of a chunk of objects into the database, and sets metadata.

        Args:
            objs: List of objects to be inserted.
            from_block: Earliest block in the current overall range.
            done_thru: Block number up to which this chunk completes.
            debug_logs: Whether debug logging is active.
        """
        if prev_task := self._db_task:
            if prev_task.done():
                if e := prev_task.exception():
                    raise e
                prev_task = None

        depth = self._depth
        self._depth += 1

        insert_coro = self.__insert_chunk(objs, from_block, done_thru, prev_task, depth, debug_logs)

        if debug_logs:
            logger._log(
                DEBUG,
                "%s queuing next db insert chunk %s thru block %s",
                (self, depth, done_thru),
            )
            task = create_task(
                coro=insert_coro,
                name=f"_insert_chunk from {from_block} to {done_thru}",
            )
        else:
            task = create_task(insert_coro)

        task._depth = depth
        self._db_task = task

    def _ensure_task(self) -> None:
        """
        Ensures there is a main fetch task running in the background. If not, creates it.
        """
        if self._task is None:
            logger.debug("creating task for %s", self)
            self._task = create_task(coro=self.__fetch(), name=f"{self}.__fetch")
            # NOTE: The task does not return and will be cancelled when this object is
            # garbage collected so there is no need to log the "destroy pending task" message.
            self._task._log_destroy_pending = False
        if self._task.done() and (e := self._task.exception()):
            # copy the exc so the traceback doesn't get destroyed by other waiters
            raise copy(e).with_traceback(e.__traceback__) from e.__cause__

    async def __insert_chunk(
        self,
        objs: List[T],
        from_block: "Block",
        done_thru: "Block",
        prev_chunk_task: Optional[Task],
        depth: int,
        debug_logs: bool,
    ) -> None:
        """
        Inserts the previously fetched chunk into the database, waits on prior tasks if needed.

        Args:
            objs: List of objects to insert.
            from_block: Earliest block for the entire fetch range.
            done_thru: Ending block for this chunk.
            prev_chunk_task: Task that inserted previous chunk, if any.
            depth: Relative ordering of this chunk within the entire fetch.
            debug_logs: Whether debug logs are active.
        """
        if prev_chunk_task:
            await prev_chunk_task
        del prev_chunk_task

        if objs:
            await self.bulk_insert(objs)
        del objs

        await _metadata_write_executor.run(self.cache.set_metadata, from_block, done_thru)
        if debug_logs:
            logger._log(
                DEBUG,
                "%s chunk %s thru block %s is now in db",
                (self, depth, done_thru),
            )

    def _prune(self, count: int) -> None:
        """
        Removes a specified number of objects from the beginning of the in-memory list.

        Args:
            count: Number of objects to remove.
        """
        self._objects = self._objects[count:]
        self._pruned += count


def _clean_addresses(addresses: Union[list, tuple]) -> Union[str, List[str]]:
    """
    Converts addresses into a standardized format, raising an error if the zero address is encountered.

    Args:
        addresses: Single or multiple addresses to clean.

    Returns:
        Cleaned addresses in a consistent string format or a list of such strings.

    Raises:
        ValueError: If the zero address is encountered or if input is invalid.
    """
    if addresses == ZERO_ADDRESS:
        raise ValueError("Cannot make a LogFilter for the zero address")
    if not addresses:
        return addresses
    if isinstance(addresses, str):
        return convert.to_address(addresses)
    elif hasattr(addresses, "__iter__"):
        if ZERO_ADDRESS in addresses:
            raise ValueError("Cannot make a LogFilter for the zero address")
        return list(map(convert.to_address, addresses))
    return convert.to_address(addresses)


def _get_suitable_checkpoint(target_block: "Block", checkpoints: Checkpoints) -> Optional["Block"]:
    """
    Finds a suitable checkpoint block that is less than or equal to a target block.

    Args:
        target_block: The block number used as a reference.
        checkpoints: Dictionary of block -> index.

    Returns:
        Most recent checkpoint block before or equal to the target block, or None if none exist.
    """
    block_lt_checkpoint, group = next(groupby(checkpoints, target_block.__lt__))
    return None if block_lt_checkpoint is True else tuple(group)[-1]


def _get_checkpoint_index(target_block: "Block", checkpoints: Checkpoints) -> Optional[int]:
    """
    Retrieves the index for a checkpoint that is less than or equal to a given block.

    Args:
        target_block: The block number reference.
        checkpoints: Dictionary of block -> index.

    Returns:
        The index of the checkpoint, or None if no suitable checkpoint exists.
    """
    checkpoint_block = _get_suitable_checkpoint(target_block, checkpoints)
    return None if checkpoint_block is None else checkpoints[checkpoint_block]



from asyncio import Task, sleep
from logging import DEBUG, Logger, StreamHandler, getLogger, _lock
from types import MethodType
from typing import Final, List, NoReturn, Optional, Tuple, TypeVar, Union, final
from weakref import WeakValueDictionary, ref as weak_ref

import a_sync
from brownie import chain
from eth_typing import BlockNumber, ChecksumAddress
from lazy_logging import LazyLoggerFactory
from typing_extensions import ParamSpec

from y.datatypes import AnyAddressType, Block
from y.networks import Network


T = TypeVar("T")
P = ParamSpec("P")


NETWORK_NAME: Final = Network.name()
yLazyLogger: Final = LazyLoggerFactory("YPRICEMAGIC")


logger: Final = getLogger(__name__)


def enable_debug_logging(logger: str = "y") -> None:
    """
    Enables ypricemagic's debugging mode. Very verbose.

    Args:
        logger: The name of the logger to enable debugging for. Defaults to "y".

    Example:
        >>> enable_debug_logging("y")
    """
    logger = getLogger(logger)
    logger.setLevel(DEBUG)
    if not logger.handlers:
        logger.addHandler(StreamHandler())


_pop_logger: Final = Logger.manager.loggerDict.pop


@final
class PriceLogger(Logger):
    enabled: bool
    address: ChecksumAddress
    block: BlockNumber
    key: Tuple[AnyAddressType, Block, Optional[str], str]
    debug_task: Optional["Task[None]"]

    def close(self) -> None:
        # since we make a lot of these we don't want logging module to cache them
        self.debug("closing %s", logger)
        with _lock:
            _pop_logger(self.name, None)


def get_price_logger(
    token_address: AnyAddressType,
    block: Block,
    *,
    symbol: str = None,
    extra: str = "",
    start_task: bool = False,
) -> PriceLogger:
    """
    Create or retrieve a `PriceLogger` instance for a given token address and block.

    This function manages a cache of loggers to ensure they have the proper members for ypricemagic.
    If a logger is enabled for DEBUG, it will start a debug task if specified.

    Args:
        token_address: The address of the token.
        block: The block number.
        symbol: An optional symbol for the token.
        extra: An optional extra string to append to the logger name.
        start_task: Whether to start a debug task. Defaults to False.

    Example:
        >>> logger = get_price_logger("0xTokenAddress", 123456)
        >>> logger.debug("This is a debug message.")

    See Also:
        - :func:`enable_debug_logging`
    """
    address = str(token_address)
    name = f"y.prices.{Network.label()}.{address}.{block}"
    if extra:
        name += f".{extra}"

    # the built-in logging module caches loggers but we need to make sure they have the proper members for ypm
    if cached_logger := _all_price_loggers.get(name, None):
        return cached_logger

    logger = getLogger(name)
    logger.address = address
    logger.block = block
    logger.enabled = logger.isEnabledFor(DEBUG)

    if logger.enabled:
        logger.debug = lambda msg, *args, **kwargs: logger._log(DEBUG, msg, args, **kwargs)

        if start_task:
            # will kill itself when this logger is garbage collected
            logger.debug_task = a_sync.create_task(
                coro=_debug_tsk(symbol, weak_ref(logger)),
                name=f"_debug_tsk({symbol}, {logger})",
                log_destroy_pending=False,
            )

    else:
        logger.debug = _noop

    logger.close = MethodType(PriceLogger.close, logger)

    _all_price_loggers[name] = logger
    return logger


def _noop(*_a, **_k): ...


async def _debug_tsk(symbol: Optional[str], logger_ref: "weak_ref[Logger]") -> NoReturn:
    """Prints a log every 1 minute until the creating coro returns."""
    if symbol:
        args = "price still fetching for %s", symbol
    else:
        args = ("still fetching...",)
    while True:
        await sleep(60)
        logger = logger_ref()
        if logger is None:
            return
        logger.debug(*args)


_all_price_loggers: Final["WeakValueDictionary[str, PriceLogger]"] = WeakValueDictionary()


NETWORK_DESCRIPTOR_FOR_ISSUE_REQ: Final = (
    f"name ({NETWORK_NAME})" if NETWORK_NAME else f"chainid ({chain.id})"
)


def _gh_issue_request(issue_request_details: Union[str, List[str]], _logger=None) -> None:
    """
    Log a request for a GitHub issue or pull request.

    Args:
        issue_request_details: The details of the issue request.
        _logger: An optional logger to use. Defaults to the module logger.

    Example:
        >>> _gh_issue_request("This is an issue request.")
    """
    _logger = _logger or logger

    if type(issue_request_details) == str:
        _logger.warning(issue_request_details)

    elif type(issue_request_details) == list:
        for message in issue_request_details:
            _logger.warning(message)

    _logger.warning(
        "Please create an issue and/or create a PR at https://github.com/BobTheBuidler/ypricemagic"
    )
    _logger.warning(
        f"In your issue, please include the network {NETWORK_DESCRIPTOR_FOR_ISSUE_REQ} and the detail shown above."
    )
    _logger.warning(
        "and I will add it soon :). This will not prevent ypricemagic from fetching price for this asset."
    )



from typing import Optional

import a_sync
from brownie import chain
from eth_typing import ChecksumAddress

from y import ENVIRONMENT_VARIABLES as ENVS
from y.datatypes import Block, UsdPrice
from y.networks import Network
from y.prices import magic

MAPPING = {
    Network.Mainnet: {
        "0x4da27a545c0c5B758a6BA100e3a049001de870f5": "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",  # stkaave -> aave
        "0x27D22A7648e955E510a40bDb058333E9190d12D4": "0x0cec1a9154ff802e7934fc916ed7ca50bde6844e",  # ppool -> pool
        # TODO: algorithmically get gauges
        "0xcF5136C67fA8A375BaBbDf13c0307EF994b5681D": "0x425BfB93370F14fF525aDb6EaEAcfE1f4e3b5802",  # sdai-usdm-gauge -> sdai-usdm
        "0x590f7e2b211Fa5Ff7840Dd3c425B543363797701": "0x5756bbdDC03DaB01a3900F01Fb15641C3bfcc457",  # YFImkUSD-gauge -> YFImkUSD
    },
}.get(chain.id, {})


def is_one_to_one_token(token_address: ChecksumAddress) -> bool:
    """
    Check if a token address is a one-to-one token.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a one-to-one token, False otherwise.

    Examples:
        >>> is_one_to_one_token("0x4da27a545c0c5B758a6BA100e3a049001de870f5")
        True

        >>> is_one_to_one_token("0x0000000000000000000000000000000000000000")
        False
    """
    return token_address in MAPPING


@a_sync.a_sync(default="sync")
async def get_price(
    token_address: ChecksumAddress,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    """
    Get the USD price of a one-to-one token by mapping it to its corresponding token.

    This function supports both synchronous and asynchronous execution.

    Args:
        token_address: The address of the token to get the price for.
        block: The block number to get the price at. Defaults to the latest block.
        skip_cache: Whether to skip the cache when fetching the price. Defaults to the value of `ENVS.SKIP_CACHE`.

    Returns:
        The USD price of the token.

    Examples:
        Synchronous usage:
        >>> get_price("0x4da27a545c0c5B758a6BA100e3a049001de870f5")

        Asynchronous usage:
        >>> await get_price("0x4da27a545c0c5B758a6BA100e3a049001de870f5", sync=False)

    See Also:
        - :func:`y.prices.magic.get_price` for the underlying price fetching logic.
    """
    return await magic.get_price(
        MAPPING[token_address], block=block, skip_cache=skip_cache, sync=False
    )



import logging
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Final, Optional, final

from brownie import Contract as BrownieContract
from brownie.exceptions import CompilerError
from eth_typing import BlockNumber, ChecksumAddress
from web3.exceptions import ContractLogicError

from y.datatypes import AnyAddressType

if TYPE_CHECKING:
    from y.prices.dex.uniswap.v2 import UniswapV2Pool


logger: Final = logging.getLogger(__name__)

# General


@final
class yPriceMagicError(ValueError):
    """
    Custom exception for ypricemagic-related errors.

    This exception is raised when any error occurs inside of ypricemagic.
    For example, if a price fails to fetch or if there are unexpected
    Exceptions while calculating prices.
    """

    def __init__(
        self,
        exc: Exception,
        token_address: ChecksumAddress,
        block: Optional[BlockNumber],
        symbol: str,
    ):
        self.token: Final = token_address
        """
        The token that caused the error.
        """

        self.block: Final = block
        """
        The block that was queried when the error occurred.
        """

        self.exception: Final = exc
        """
        The original :class:`~Exception` that was raised and wrapped with the yPriceMagicError.
        """

        detail = exc.__class__.__name__
        if not isinstance(self.exception, PriceError):
            detail += f"({exc})"
        # We do the import here to avoid a circular dependency
        from y import Network

        super().__init__(
            f"{detail} while fetching {Network.printable()} {symbol} {token_address} at block {block}"
        )


@final
class PriceError(Exception):
    """
    Raised when a queried price is not found.
    """

    def __init__(self, logger: logging.Logger, symbol: str):
        super().__init__(f"No price found for {symbol} {logger.address} at block {logger.block}")


@final
class UnsupportedNetwork(ValueError):
    """
    Raised when an operation is attempted on an unsupported blockchain network.
    """


@final
class NonStandardERC20(Exception):
    """
    Raised when an ERC20 token contract is expected but the provided address is not a standard ERC20 token.
    """


@final
class CantFetchParam(Exception):
    pass


@final
class TokenError(ValueError):
    """Raised when a token contract is not the correct contract type for the desired operation."""

    def __init__(self, token: AnyAddressType, desired_type: str, *optional_extra_args: Any):
        super().__init__(f"{token} is not a {desired_type}", *optional_extra_args)


# Explorer Exceptions


class _ExplorerError(Exception):
    """
    Base class for exceptions related to block explorer interactions.

    These errors are specific to issues encountered when interacting with blockchain explorers like Etherscan.
    """


@final
class InvalidAPIKeyError(_ExplorerError):
    """
    Raised when the API key for the block explorer has been rejected.
    This typically occurs when making requests to a block explorer API with a missing, incorrect, or banned key.
    """

    _msg: Final = "The block explorer for this network says your API key is invalid."

    def __init__(self, msg: str = ""):
        super().__init__(msg or self._msg)


# Contracts


@final
class ContractNotVerified(_ExplorerError):
    """
    Raised when attempting to fetch the ABI for an unverified contract from a block explorer.
    """


@final
class NoProxyImplementation(Exception):
    """
    Raised when the implementation address of a proxy contract cannot be determined.

    This may occur when trying to interact with proxy contracts that don't follow standard patterns.
    """


@final
class MessedUpBrownieContract(Exception):
    """
    Raised when there's an issue initialized a Brownie contract instance,
    typically in the compilation step.
    """

    def __init__(self, address: ChecksumAddress, *args: object) -> None:
        # sourcery skip: remove-pass-elif
        super().__init__(*args)
        # try to recache the contract
        try:
            BrownieContract.from_explorer(address)
        except CompilerError:
            pass  # didn't work, oh well
        except Exception as e:
            if "invalid literal for int() with base 16: ''" in str(e):
                pass  # didn't work, oh well
            elif "list index out of range" in str(e):
                pass  # didn't work, oh well
            elif "pop from an empty deque" in str(e):
                pass  # didn't work, oh well
            elif "'UsingForDirective' object has no attribute 'typeName'" in str(e):
                pass  # didn't work, oh well
            elif contract_not_verified(e):
                pass  # not verified, won't work
            else:
                raise


def contract_not_verified(e: Exception) -> bool:
    triggers = (
        "Contract source code not verified",
        "has not been verified",
    )
    stre = str(e)
    for trigger in triggers:
        if trigger in stre:
            return True
    return False


# Pool detection


@final
class NotAUniswapV2Pool(Exception):
    """
    Raised when a contract is incorrectly identified as a Uniswap V2 pool.
    """

    # NOTE: we use this exc to get the non-pool out of the pool singleton so it doesn't grow forever
    # TODO: Refactor this goofy thing out
    def __init__(self, non_pool: "UniswapV2Pool"):
        from y.prices.dex.uniswap.v2 import UniswapV2Pool

        UniswapV2Pool._ChecksumASyncSingletonMeta__instances[True].pop(non_pool.address, None)
        UniswapV2Pool._ChecksumASyncSingletonMeta__instances[False].pop(non_pool.address, None)
        super().__init__(non_pool.address)


@final
class NotABalancerV2Pool(Exception):
    """
    Raised when a contract is incorrectly identified as a Balancer V2 pool.
    """


# Uni


@final
class CantFindSwapPath(Exception):
    pass


@final
class TokenNotFound(ValueError):
    """
    Raised when a specified token cannot be found in a given container.

    This is usually used when searching for a token in a liquidity pool.
    """

    def __init__(self, token, container):
        super().__init__(f"{token} is not in {container}")


# Calls


@final
class CalldataPreparationError(Exception):
    """
    Raised when there's an error in preparing calldata for a contract interaction.
    """


@final
class CallReverted(Exception):
    """
    Raised when a contract call is reverted.
    """


def call_reverted(e: Exception) -> bool:
    if isinstance(e, ContractLogicError):
        return True
    triggers = (
        "execution reverted",
        "No data was returned - the call likely reverted",
        "invalid opcode: opcode 0xfe not defined",
        "Tried to read 32 bytes.  Only got 0 bytes",
        "error processing call Revert",
        "invalid opcode: INVALID",
    )
    stre = str(e)
    for trigger in triggers:
        if trigger in stre:
            return True
    return False


def continue_if_call_reverted(e: Exception) -> None:
    if call_reverted(e):
        return
    else:
        raise e


def out_of_gas(e: Exception) -> bool:
    return "out of gas" in str(e)


# Provider Exceptions:


@final
class NodeNotSynced(Exception):
    pass


@contextmanager
def reraise_excs_with_extra_context(*extra_context: Any, after: bool = True):
    try:
        yield
    except Exception as e:
        e.args = (*e.args, *extra_context)
        raise



from typing import List, Optional, Tuple

from a_sync import cgather

from y._decorators import continue_on_revert, stuck_coro_debugger
from y.datatypes import Address, Block
from y.exceptions import call_reverted
from y.prices.dex.uniswap.v2 import Path, UniswapRouterV2, UniswapV2Pool
from y.utils.cache import a_sync_ttl_cache


class SolidlyRouterBase(UniswapRouterV2):
    """
    Solidly is a modified fork of Uni V2.
    The `uniswap_multiplexer` is the entrypoint for pricing using this object.

    This class provides methods to interact with the Solidly protocol, which is
    based on the Uniswap V2 model. It includes functionality to get price quotes
    and determine routes for token swaps.

    See Also:
        - :class:`~y.prices.dex.uniswap.v2.UniswapRouterV2`
        - :class:`~y.prices.dex.uniswap.v2.UniswapV2Pool`
    """

    @continue_on_revert
    @stuck_coro_debugger
    async def get_quote(
        self, amount_in: int, path: Path, block: Optional[Block] = None
    ) -> Tuple[int, int]:
        """
        Get a price quote for a given input amount and swap path.

        This method calculates the output amount for a given input amount and
        swap path by interacting with the Solidly contract.

        Args:
            amount_in: The amount of input tokens.
            path: The swap path as a list of token addresses.
            block: The block number to query. Defaults to the latest block.

        Returns:
            A tuple containing the output amount and the path used.

        Raises:
            Exception: If the call reverts for reasons other than insufficient
            input amount or liquidity.

        Examples:
            >>> router = SolidlyRouterBase("0xRouterAddress")
            >>> quote = await router.get_quote(1000, ["0xTokenA", "0xTokenB"])
            >>> print(quote)
        """
        routes = await self.get_routes_from_path(path, block)
        try:
            return await self.contract.getAmountsOut.coroutine(
                amount_in, routes, block_identifier=block
            )
        except Exception as e:
            strings = (
                "INSUFFICIENT_INPUT_AMOUNT",
                "INSUFFICIENT_LIQUIDITY",
                "INSUFFICIENT_OUT_LIQUIDITY",
                "Sequence has incorrect length",
                "Call reverted: Integer overflow",
            )
            if not call_reverted(e) and not any(map(str(e).__contains__, strings)):
                raise


class SolidlyPool(UniswapV2Pool):
    """
    Represents a liquidity pool in the Solidly protocol.

    This class inherits from :class:`~y.prices.dex.uniswap.v2.UniswapV2Pool` and
    provides the same interface for interacting with liquidity pools.

    See Also:
        - :class:`~y.prices.dex.uniswap.v2.UniswapV2Pool`
    """

    pass


class SolidlyRouter(SolidlyRouterBase):
    """
    A router for interacting with the Solidly protocol.

    This class extends :class:`~SolidlyRouterBase` to provide additional
    functionality specific to the Solidly protocol, such as determining the
    appropriate pool for a token pair and calculating swap routes.

    See Also:
        - :class:`~SolidlyRouterBase`
    """

    @stuck_coro_debugger
    @a_sync_ttl_cache
    async def pair_for(self, input_token: Address, output_token: Address, stable: bool) -> Address:
        """
        Get the address of the pool for a given token pair.

        This method returns the address of the pool that contains the specified
        input and output tokens, and indicates whether the pool is stable.

        Args:
            input_token: The address of the input token.
            output_token: The address of the output token.
            stable: A boolean indicating whether to look for a stable pool.

        Returns:
            The address of the pool.

        Examples:
            >>> router = SolidlyRouter("0xRouterAddress")
            >>> pool_address = await router.pair_for("0xTokenA", "0xTokenB", True)
            >>> print(pool_address)
        """
        return await self.contract.pairFor.coroutine(input_token, output_token, stable)

    @stuck_coro_debugger
    @a_sync_ttl_cache
    async def get_pool(
        self, input_token: Address, output_token: Address, stable: bool, block: Block
    ) -> Optional[SolidlyPool]:
        """
        Get the pool object for a given token pair.

        This method returns a :class:`~SolidlyPool` object representing the pool
        that contains the specified input and output tokens, and indicates
        whether the pool is stable.

        Args:
            input_token: The address of the input token.
            output_token: The address of the output token.
            stable: A boolean indicating whether to look for a stable pool.
            block: The block number to query.

        Returns:
            A :class:`~SolidlyPool` object if the pool exists, otherwise None.

        Examples:
            >>> router = SolidlyRouter("0xRouterAddress")
            >>> pool = await router.get_pool("0xTokenA", "0xTokenB", True, 12345678)
            >>> print(pool)
        """
        pool_address = await self.pair_for(input_token, output_token, stable, sync=False)
        if await self.contract.isPair.coroutine(pool_address, block_identifier=block):
            return SolidlyPool(pool_address, asynchronous=self.asynchronous)

    @stuck_coro_debugger
    async def get_routes_from_path(
        self, path: Path, block: Block
    ) -> List[Tuple[Address, Address, bool]]:
        """
        Determine the swap routes from a given path.

        This method calculates the swap routes for a given path by checking for
        available stable and unstable pools and selecting the deepest pool.

        Args:
            path: The swap path as a list of token addresses.
            block: The block number to query.

        Returns:
            A list of tuples, each containing the input token, output token, and
            a boolean indicating whether the pool is stable.

        Raises:
            ValueError: If no pool is found for a token pair.

        Examples:
            >>> router = SolidlyRouter("0xRouterAddress")
            >>> routes = await router.get_routes_from_path(["0xTokenA", "0xTokenB"], 12345678)
            >>> print(routes)
        """
        routes = []
        for i in range(len(path) - 1):
            input_token, output_token = path[i], path[i + 1]
            # Try for a stable pool first and use that if available
            stable_pool, unstable_pool = await cgather(
                self.get_pool(input_token, output_token, True, block, sync=False),
                self.get_pool(input_token, output_token, False, block, sync=False),
            )

            if stable_pool and unstable_pool:
                # We have to find out which of these pools is deepest
                stable_reserves, unstable_reserves = await cgather(
                    stable_pool.reserves(block=block, sync=False),
                    unstable_pool.reserves(block=block, sync=False),
                )
                stable_reserves = tuple(stable_reserves)
                unstable_reserves = tuple(unstable_reserves)

                # NOTE: using `__token0__` and `__token1__` is faster than `__tokens__` since they're already cached and return instantly
                #       it also creates 2 fewer tasks and 1 fewer future than `__tokens__` since there is no use of `asyncio.gather`.
                if (
                    await stable_pool.__token0__ == await unstable_pool.__token0__
                    and await stable_pool.__token1__ == await unstable_pool.__token1__
                ):
                    stable_reserve = stable_reserves[0]
                    unstable_reserve = unstable_reserves[0]
                else:  # Order of tokens is flip flopped in the pools
                    stable_reserve = stable_reserves[0]
                    unstable_reserve = unstable_reserves[1]
                if stable_reserve >= unstable_reserve:
                    is_stable = True
                elif stable_reserve < unstable_reserve:
                    is_stable = False
                routes.append([input_token, output_token, is_stable])
            elif stable_pool:
                routes.append([input_token, output_token, True])
            elif unstable_pool:
                routes.append([input_token, output_token, False])
            else:
                raise ValueError("Not sure why this function is even running if no pool is found")

        return routes



from decimal import Decimal
from typing import Tuple

from a_sync import a_sync, cgather
from async_lru import alru_cache

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20
from y.contracts import Contract, has_method, is_contract
from y.datatypes import Address, Block
from y.exceptions import ContractNotVerified


try:
    oracle = "0x9a9Fa8338dd5E5B2188006f1Cd2Ef26d921650C2"
    PENDLE_ORACLE = Contract(oracle) if is_contract(oracle) else None
except ContractNotVerified:
    PENDLE_ORACLE = None

TWAP_DURATION = 1


@a_sync("sync")
async def is_pendle_lp(token: Address) -> bool:
    """
    Determines if the given token address is a Pendle LP token.

    Args:
        token: The address of the token to check.

    Returns:
        True if the token is a Pendle LP token, False otherwise.

    Example:
        >>> is_pendle = is_pendle_lp("0x1b92b5242301ce4a8c73cc3ef0d6dee33a3a5b23")
        >>> print(is_pendle)
        True
    """
    return PENDLE_ORACLE is not None and await has_method(
        token, "readTokens()(address,address,address)", sync=False
    )


@alru_cache(maxsize=None)
async def get_tokens(lp_token: Address) -> Tuple[str, str, str]:
    """
    Retrieves the addresses of the tokens in a Pendle LP token.

    This function is cached to improve performance for repeated calls.

    Args:
        lp_token: The address of the Pendle LP token.

    Returns:
        A tuple containing the addresses of the SY token, PT token, and YT token.

    Example:
        >>> tokens = await get_tokens("0x1b92b5242301ce4a8c73cc3ef0d6dee33a3a5b23")
        >>> print(tokens)
        ('0x...', '0x...', '0x...')  # Addresses of SY, PT, and YT tokens

    Note:
        The function returns the result of the `readTokens` method from the Pendle LP contract, which provides the addresses of the underlying tokens.

    See Also:
        - :func:`is_pendle_lp` for checking if a token is a Pendle LP token.
    """
    lp_token = await Contract.coroutine(lp_token)
    return await lp_token.readTokens


@a_sync("sync")
async def get_lp_price(
    token: Address, block: Block = None, skip_cache: bool = ENVS.SKIP_CACHE
) -> Decimal:
    """
    Calculates the price of a Pendle LP token.

    Args:
        token: The address of the Pendle LP token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching prices. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the LP token in USD.

    Example:
        >>> price = get_lp_price("0x1b92b5242301ce4a8c73cc3ef0d6dee33a3a5b23", block=14_000_000)
        >>> print(f"{price:.6f}")
        1.234567  # The price of the Pendle LP token in USD

    Note:
        This function retrieves the LP to asset rate using the Pendle Oracle contract and calculates the LP token price by multiplying the LP to asset rate by the asset's price in USD.

    See Also:
        - :func:`get_tokens` for retrieving the tokens in a Pendle LP token.
    """
    tokens = await get_tokens(str(token))  # force to string for cache key
    # NOTE: we might not need this, leave it commented out for now
    # names = await cgather(*(ERC20(t, asynchronous=True).name for t in tokens))
    # if any("DAI" in name for name in names):
    #    use_asset = True
    #    rate = await PENDLE_ORACLE.getLpToAssetRate.coroutine(token, twap_duration, block_identifier=block)
    # elif any("crvUSD" in name for name in (ERC20(t) for t in tokens)):
    #    use_asset = True
    #    rate = await PENDLE_ORACLE.getLpToAssetRate.coroutine(token, twap_duration, block_identifier=block)
    # else:
    #    use_asset = False
    #    rate = await PENDLE_ORACLE.getLpToAssetRate.coroutine(token, twap_duration, block_identifier=block)
    sy_token, p_token, y_token = tokens
    sy, rate = await cgather(
        Contract.coroutine(sy_token),
        PENDLE_ORACLE.getLpToAssetRate.coroutine(token, TWAP_DURATION, block_identifier=block),
    )
    _, asset, decimals = await sy.assetInfo
    rate /= Decimal(10**decimals)
    return rate * await ERC20(asset, asynchronous=True).price(block=block, skip_cache=skip_cache)
